#
# review_wdgt.py <Peter.Bienstman@UGent.be>
#

import _main_widget
import _review_widget

from mnemosyne.libmnemosyne.ui_components.review_widget import ReviewWidget


class ReviewWdgt(ReviewWidget):

    def redraw_now(self):
        pass

    def empty(self):
        background = "white"
        if self.review_controller().card:
            colour = self.config().card_type_property(\
            "background_colour", self.review_controller().card.card_type)
            if colour:
                background = ("%X" % colour)[2:] # Strip alpha.
        return """
        <html><head>
        <style type="text/css">
        table { height: 100%; }
        body  { background-color: """ + background + """;
                margin: 0;
                padding: 0;
                border: thin solid #8F8F8F; }
        </style></head>
        <body><table><tr><td>
        </td></tr></table></body></html>"""

    def show_answer(self):
        self.review_controller().show_answer()

    def grade_answer(self, grade):
        self.review_controller().grade_answer(grade)

    def set_question_box_visible(self, is_visible):
        _review_widget.set_question_box_visible(is_visible)

    def set_answer_box_visible(self, is_visible):
        _review_widget.set_answer_box_visible(is_visible)

    def set_question_label(self, text):
        _review_widget.set_question_label(text)

    def set_question(self, text):
        self.question = text

    def set_answer(self, text):
        self.answer = text

    def reveal_question(self):
        _review_widget.set_question(self.question)

    def reveal_answer(self, process_audio=True):
        _review_widget.set_answer(self.answer, process_audio)

    def clear_question(self):
        self.question = self.empty()
        self.reveal_question()

    def clear_answer(self):
        # We don't process the audio here, as that would kill the pending
        # audio events from the question.
        self.answer = self.empty()
        self.reveal_answer(process_audio=False)

    def update_show_button(self, text, is_default, is_enabled):
        _review_widget.update_show_button(text, is_default, is_enabled)

    def set_grades_enabled(self, is_enabled):
        _review_widget.set_grades_enabled(is_enabled)

    def set_default_grade(self, grade):
        pass

    def update_status_bar_counters(self):
        scheduled_count, non_memorised_count, active_count = \
            self.review_controller().counters()
        counters = "Sch.: %d Not mem.: %d Act.: %d" % \
                    (scheduled_count, non_memorised_count, active_count)
        _main_widget.set_status_bar_message(counters)




#
# sync_dlg.py <Peter.Bienstman@UGent.be>
#

import _dialogs

from mnemosyne.libmnemosyne.ui_components.dialogs import SyncDialog


class SyncDlg(SyncDialog):

    def __init__(self, component_manager):
        SyncDialog.__init__(self, component_manager)

    def activate(self):
        SyncDialog.activate(self)
        _dialogs.sync_dlg_activate()



# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import os, sys, re
import gettext
import threading

langs = [
    ("Afrikaans", "af"),
    ("Bahasa Melayu", "ms"),
    ("Dansk", "da"),
    ("Deutsch", "de"),
    ("Eesti", "et"),
    ("English", "en"),
    ("Español", "es"),
    ("Esperanto", "eo"),
    ("Euskara", "eu"),
    ("Français", "fr"),
    ("Galego", "gl"),
    ("Hrvatski", "hr"),
    ("Interlingua", "ia"),
    ("Italiano", "it"),
    ("Lenga d'òc", "oc"),
    ("Magyar", "hu"),
    ("Nederlands","nl"),
    ("Norsk","nb"),
    ("Occitan","oc"),
    ("Plattdüütsch", "nds"),
    ("Polski", "pl"),
    ("Português Brasileiro", "pt_BR"),
    ("Português", "pt"),
    ("Româneşte", "ro"),
    ("Slovenčina", "sk"),
    ("Slovenščina", "sl"),
    ("Suomi", "fi"),
    ("Svenska", "sv"),
    ("Tiếng Việt", "vi"),
    ("Türkçe", "tr"),
    ("Čeština", "cs"),
    ("Ελληνικά", "el"),
    ("Ελληνικά", "el"),
    ("босански", "bs"),
    ("Български", "bg"),
    ("Монгол хэл","mn"),
    ("русский язык", "ru"),
    ("Српски", "sr"),
    ("українська мова", "uk"),
    ("עִבְרִית", "he"),
    ("العربية", "ar"),
    ("فارسی", "fa"),
    ("ภาษาไทย", "th"),
    ("日本語", "ja"),
    ("简体中文", "zh_CN"),
    ("繁體中文", "zh_TW"),
    ("한국어", "ko"),
]

threadLocal = threading.local()

# global defaults
currentLang = None
currentTranslation = None

def localTranslation():
    "Return the translation local to this thread, or the default."
    if getattr(threadLocal, 'currentTranslation', None):
        return threadLocal.currentTranslation
    else:
        return currentTranslation

def _(str):
    return localTranslation().gettext(str)

def ngettext(single, plural, n):
    return localTranslation().ngettext(single, plural, n)

def langDir():
    dir = os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "locale")
    if not os.path.isdir(dir):
        try:
            dir = os.path.join(os.path.dirname(sys.argv[0]), "locale")
        except: # MNEMOSYNE: android does not have sys.argv
            dir = ""
    if not os.path.isdir(dir):
        dir = "/usr/share/anki/locale"
    return dir

def setLang(lang, local=True):
    trans = gettext.translation(
        'anki', langDir(), languages=[lang], fallback=True)
    if local:
        threadLocal.currentLang = lang
        threadLocal.currentTranslation = trans
    else:
        global currentLang, currentTranslation
        currentLang = lang
        currentTranslation = trans

def getLang():
    "Return the language local to this thread, or the default."
    if getattr(threadLocal, 'currentLang', None):
        return threadLocal.currentLang
    else:
        return currentLang

def noHint(str):
    "Remove translation hint from end of string."
    return re.sub("(^.*?)( ?\(.+?\))?$", "\\1", str)

if not currentTranslation:
    setLang("en_US", local=False)



#
# qwebengineview2.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets, QtWebEngineWidgets

from mnemosyne.libmnemosyne.gui_translator import _


class QWebEngineView2(QtWebEngineWidgets.QWebEngineView):

    """QWebEngineView which restores the focus to the review widget,
    so that the keyboard shortcuts still continue to work.

    """

    def __init__(self, parent=None):
        super().__init__(parent)
        QtWebEngineWidgets.QWebEngineProfile.defaultProfile().\
         setHttpCacheType(QtWebEngineWidgets.QWebEngineProfile.MemoryHttpCache)
        #self.page().profile().setPersistentCookiesPolicy(\
        #    QtWebEngineWidgets.QWebEngineProfile.NoPersistentCookies)
        # self.linkClicked.connect(self.link_clicked)
        # self.page().setLinkDelegationPolicy(QtWebKit.QWebPage.DelegateAllLinks)

    def link_clicked(self, url):
        import webbrowser  # Slow import
        webbrowser.open(url.toString())




#
# test_convert_cards.py <Peter.Bienstman@UGent.be>
#

import os
import sys
import copy
import shutil

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne import Mnemosyne
from mnemosyne.libmnemosyne.card_type_converter import CardTypeConverter
from mnemosyne.libmnemosyne.ui_components.main_widget import MainWidget

class Widget(MainWidget):

    def show_question(self, question, option0, option1, option2):
        if question.startswith("This will delete cards and their history"):
            return 0  # Proceed and delete":
        if question.startswith("Can't preserve history when converting"):
            return 0  # Reset learning history
        if question.startswith("Identical card is already in database"):
            return  0 # Do not add
        raise NotImplementedError


class TestConverter:

    def test(self):
        c = CardTypeConverter(None)
        new_cards, edited_cards, deleted_cards \
                   = c.convert(None, None, None, None)
        assert new_cards == []
        assert edited_cards == []
        assert deleted_cards == []

class TestConvertCards(MnemosyneTest):

    def setup(self):
        self.initialise_data_dir()
        path = os.path.join(os.getcwd(), "..", "mnemosyne", "libmnemosyne",
                            "renderers")
        if path not in sys.path:
            sys.path.append(path)
        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_convert_cards", "Widget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"),  automatic_upgrades=False)
        self.review_controller().reset()

        from mnemosyne.libmnemosyne.card_types.map import MapPlugin
        for plugin in self.plugins():
            if isinstance(plugin, MapPlugin):
                plugin.activate()
                break

    def test_1_to_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        old_card = copy.copy(self.database().cards_from_fact(fact)[0])

        new_fact_data = {"f": "question2",
                         "b": "answer2"}
        new_card_type = self.card_type_with_id("2")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"], correspondence=[])

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)

        assert new_card_1.card_type.id == "2"
        assert new_card_2.card_type.id == "2"

        if new_card_1.fact_view.id == "2.1":
            print(new_card_1.id, new_card_2.id, old_card.id)
            assert new_card_1 == old_card
            assert new_card_2 != old_card
            assert new_card_2.grade == -1
        else:
            assert new_card_2 == old_card
            assert new_card_1 != old_card
            assert new_card_1.grade == -1

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_1_to_2_multi(self):
        open("a.ogg", "w")
        full_path = os.path.abspath("a.ogg")

        fact_data = {"f": "<img src=\"%s\">" % full_path,
                     "b": "answer"}

        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        old_card = copy.copy(self.database().cards_from_fact(fact)[0])

        new_fact_data = {"f": "question2",
                         "b": "answer2"}
        new_card_type = self.card_type_with_id("2")
        self.controller().change_card_type([fact], card.card_type,
               new_card_type, correspondence=[])

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)

        assert new_card_1.card_type.id == "2"
        assert new_card_2.card_type.id == "2"

        if new_card_1.fact_view.id == "2.1":
            assert new_card_1 == old_card
            assert new_card_2 != old_card
            assert new_card_2.grade == -1
        else:
            assert new_card_2 == old_card
            assert new_card_1 != old_card
            assert new_card_1.grade == -1

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_1_to_2_multi_no_media(self):
        fact_data = {"f": "question",
                     "b": "answer"}

        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        old_card = copy.copy(self.database().cards_from_fact(fact)[0])

        new_fact_data = {"f": "question2",
                         "b": "answer2"}
        new_card_type = self.card_type_with_id("2")
        self.controller().change_card_type([fact], card.card_type,
               new_card_type, correspondence=[])

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)

        assert new_card_1.card_type.id == "2"
        assert new_card_2.card_type.id == "2"

        if new_card_1.fact_view.id == "2.1":
            assert new_card_1 == old_card
            assert new_card_2 != old_card
            assert new_card_2.grade == -1
        else:
            assert new_card_2 == old_card
            assert new_card_1 != old_card
            assert new_card_1.grade == -1

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_2_to_1(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("2")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question2",
                         "b": "answer2"}
        new_card_type = self.card_type_with_id("1")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"], correspondence=[])

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(fact)[0]
        assert new_card.card_type.id == "1"

        if old_card_1.fact_view.id == "2.1":
            assert new_card == old_card_1
            assert new_card != old_card_2
        else:
            assert new_card == old_card_2
            assert new_card != old_card_1

        new_card.question()
        new_card.answer()

    def test_1_to_3_a(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        old_card = copy.copy(self.database().cards_from_fact(fact)[0])

        new_fact_data = {"f": "foreign word",
                         "p_1": "pronunciation",
                         "m_1": "translation"}
        correspondence = {"f": "f", "b": "p_1"}
        new_card_type = self.card_type_with_id("3")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)
        assert new_card_1.card_type.id == "3"
        assert new_card_2.card_type.id == "3"

        if new_card_1.fact_view.id == "3.1":
            assert new_card_1 == old_card
            assert new_card_2 != old_card
        else:
            assert new_card_2 == old_card
            assert new_card_1 != old_card

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_1_to_3_b(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        old_card = copy.copy(self.database().cards_from_fact(fact)[0])

        new_fact_data = {"f": "foreign word",
                         "p_1": "pronunciation",
                         "m_1": "translation"}
        correspondence = {"f": "m_1", "b": "f"}
        new_card_type = self.card_type_with_id("3")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)
        assert new_card_1.card_type.id == "3"
        assert new_card_2.card_type.id == "3"

        if new_card_1.fact_view.id == "3.2":
            assert new_card_1 == old_card
            assert new_card_2 != old_card
        else:
            assert new_card_2 == old_card
            assert new_card_1 != old_card

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_3_to_1_a(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "f", "m_1": "b"}
        new_card_type = self.card_type_with_id("1")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(fact)[0]
        assert new_card.card_type.id == "1"

        if old_card_1.fact_view.id == "3.1":
            assert new_card == old_card_1
            assert new_card != old_card_2
        else:
            assert new_card == old_card_2
            assert new_card != old_card_1

        new_card.question()
        new_card.answer()

    def test_3_to_1_b(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "b", "m_1": "f"}
        new_card_type = self.card_type_with_id("1")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(fact)[0]
        assert new_card.card_type.id == "1"

        if old_card_1.fact_view.id == "3.2":
            assert new_card == old_card_1
            assert new_card != old_card_2
        else:
            assert new_card == old_card_2
            assert new_card != old_card_1

        new_card.question()
        new_card.answer()

    def test_3_to_1_c(self):
        # Missing required field.
        fact_data = {"f": "foreign word",
                     "m_1": "meaning"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"b": "foreign word"}
        correspondence = {"f": "b", "p_1": "f"}
        new_card_type = self.card_type_with_id("1")
        self.controller().change_card_type([fact], card_type,
               new_card_type, correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        card_1, card_2 = self.database().cards_from_fact(fact)
        assert card_1.card_type.id == "3"
        assert card_2.card_type.id == "3"


    def test_2_to_3_a(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("2")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "foreign word",
                         "p_1": "pronunciation",
                         "m_1": "translation"}
        correspondence = {"f": "f", "b": "m_1"}
        new_card_type = self.card_type_with_id("3")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)
        assert new_card_1.card_type.id == "3"
        assert new_card_2.card_type.id == "3"

        for old in [old_card_1, old_card_2]:
            for new in [new_card_1, new_card_2]:
                if old == new:
                    assert old.fact_view.id.split(".")[1] == \
                           new.fact_view.id.split(".")[1]

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_2_to_3_b(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("2")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "foreign word",
                         "p_1": "pronunciation",
                         "m_1": "translation"}
        correspondence = {"f": "m_1", "b": "f"}
        new_card_type = self.card_type_with_id("3")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)
        assert new_card_1.card_type.id == "3"
        assert new_card_2.card_type.id == "3"

        for old in [old_card_1, old_card_2]:
            for new in [new_card_1, new_card_2]:
                if old == new:
                    assert old.fact_view != new.fact_view

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_3_to_2_a(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "f", "m_1": "b"}
        new_card_type = self.card_type_with_id("2")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)

        assert new_card_1.card_type.id == "2"
        assert new_card_2.card_type.id == "2"

        for old in [old_card_1, old_card_2]:
            for new in [new_card_1, new_card_2]:
                if old == new:
                    assert old.fact_view.id.split(".")[1] == \
                           new.fact_view.id.split(".")[1]

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_3_to_2_b(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "b", "m_1": "f"}
        new_card_type = self.card_type_with_id("2")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card_1, new_card_2 = self.database().cards_from_fact(fact)

        assert new_card_1.card_type.id == "2"
        assert new_card_2.card_type.id == "2"

        for old in [old_card_1, old_card_2]:
            for new in [new_card_1, new_card_2]:
                if old == new:
                    assert old.fact_view != new.fact_view

        new_card_1.question()
        new_card_1.answer()
        new_card_2.question()
        new_card_2.answer()

    def test_3_clone_to_1_a(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        self.controller().clone_card_type(card_type, "my_3")
        card_type = self.card_type_with_id("3::my_3")

        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "f", "m_1": "b"}
        new_card_type = self.card_type_with_id("1")
        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(fact)[0]
        assert new_card.card_type.id == "1"

        if old_card_1.fact_view.id == "3::my_3.1":
            assert new_card == old_card_1
            assert new_card != old_card_2
        else:
            assert new_card == old_card_2
            assert new_card != old_card_1

        new_card.question()
        new_card.answer()

    def test_3_to_1_clone_a(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "f", "m_1": "b"}
        new_card_type = self.card_type_with_id("1")
        self.controller().clone_card_type(new_card_type, "my_1")
        new_card_type = self.card_type_with_id("1::my_1")

        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(fact)[0]
        assert new_card.card_type.id == "1::my_1"
        assert new_card.fact.data["f"] == "question"
        assert new_card.fact.data["b"] == "answer"

        if old_card_1.fact_view.id == "3.1":
            assert new_card == old_card_1
            assert new_card != old_card_2
        else:
            assert new_card == old_card_2
            assert new_card != old_card_1

        new_card.question()
        new_card.answer()

    def test_3_clone_to_1_clone_a(self):
        fact_data = {"f": "foreign word",
                     "p_1": "pronunciation",
                     "m_1": "translation"}
        card_type = self.card_type_with_id("3")
        self.controller().clone_card_type(card_type, "my_3")
        card_type = self.card_type_with_id("3::my_3")

        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]

        fact = card.fact
        card_1, card_2 = self.database().cards_from_fact(fact)
        old_card_1 = copy.copy(card_1)
        old_card_2 = copy.copy(card_2)

        new_fact_data = {"f": "question",
                         "b": "answer"}
        correspondence = {"f": "f", "m_1": "b"}
        new_card_type = self.card_type_with_id("1")
        self.controller().clone_card_type(new_card_type, "my_1")
        new_card_type = self.card_type_with_id("1::my_1")

        self.controller().edit_card_and_sisters(card, new_fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence=correspondence)

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(fact)[0]
        assert new_card.card_type.id == "1::my_1"

        if old_card_1.fact_view.id == "3::my_3.1":
            assert new_card == old_card_1
            assert new_card != old_card_2
        else:
            assert new_card == old_card_2
            assert new_card != old_card_1

        new_card.question()
        new_card.answer()

    def test_cloze_to_1(self):
        from mnemosyne.libmnemosyne.ui_components.statistics_widget import \
             StatisticsWidget
        from mnemosyne.libmnemosyne.statistics_pages.schedule import Schedule
        class ScheduleWdgt(StatisticsWidget):
            used_for = Schedule
        self.mnemosyne.component_manager.register(ScheduleWdgt)

        for plugin in self.plugins():
            component = plugin.components[0]
            if component.component_type == "card_type" and component.id == "5":
                plugin.activate()

        fact_data = {"text": "[question]"}
        card_type = self.card_type_with_id("5")
        card = self.controller().create_new_cards(fact_data, card_type,
                                          grade=-1, tag_names=["default"])[0]
        self.review_controller().reset()

        new_card_type = self.card_type_with_id("1")
        fact_data = {"f": "[question]", "b": ""}
        self.controller().edit_card_and_sisters(card, fact_data,
               new_card_type, new_tag_names=["default2"],
               correspondence={'text': "f"})

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 1

        new_card = self.database().cards_from_fact(card.fact)[0]

        new_card.question()
        new_card.answer()

    def test_2_to_2_clone(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("2")
        card_1, card_2  = self.controller().create_new_cards(fact_data,
                                 card_type, grade=-1, tag_names=["default"])

        new_card_type = self.controller().\
                        clone_card_type(card_type, "my_2")

        self.controller().edit_card_and_sisters(card_1, fact_data,
               new_card_type, new_tag_names=["default"],
               correspondence={})

        assert self.database().fact_count() == 1
        assert self.database().card_count() == 2

        new_card = self.database().cards_from_fact(card_1.fact)[0]
        assert new_card.card_type.id == "2::my_2"
        assert new_card.fact.data["f"] == "question"
        assert new_card.fact.data["b"] == "answer"

        new_card.question()
        new_card.answer()

    def test_1_to_3_clone(self):
        card_type = self.card_type_with_id("1") # Production only.
        fact_data = {"f": "translation",
                     "b": "foreign"}
        card = self.controller().create_new_cards(fact_data,
            card_type, grade=2, tag_names=["default"])[0]

        new_card_type = self.controller().\
                        clone_card_type(self.card_type_with_id("3"), "my_language")
        correspondence = {"f": "m_1", "b": "f"}
        self.controller().change_card_type([card.fact], card_type, new_card_type,
                         correspondence=correspondence)

        fact = self.database().fact(card.fact._id, is_id_internal=True)
        assert fact["f"] == "foreign"
        card_1, card_2 = self.database().cards_from_fact(fact)

        assert len(card_1.tags) == 1
        assert len(card_2.tags) == 1
        for tag in card_1.tags:
            assert tag.name == "default"
        for tag in card_2.tags:
            assert tag.name == "default"

        assert self.database().con.execute("select tags from cards where _id=?",
            (card_1._id, )).fetchone()[0] == "default"
        assert self.database().con.execute("select tags from cards where _id=?",
            (card_2._id, )).fetchone()[0] == "default"

        assert card_1.active == True
        assert card_2.active == True

        if card_1.fact_view.id == "3.1": # Recognition
            assert "foreign" in  card_1.question()
            assert "translation" in card_2.question()
            assert card_1.grade == -1
            assert card_2.grade == 2
        else:
            assert "foreign" in  card_2.question()
            assert "translation" in card_1.question()
            assert card_2.grade == -1
            assert card_1.grade == 2

    def test_1_to_3_clone_bis(self):
        card_type = self.card_type_with_id("1") # Production only.
        fact_data = {"f": "translation",
                     "b": "foreign"}
        card = self.controller().create_new_cards(fact_data,
            card_type, grade=2, tag_names=["default"])[0]

        new_card_type = self.controller().\
                        clone_card_type(self.card_type_with_id("3"), "my_language")
        correspondence = {"f": "m_1", "b": "f"}

        new_fact_data = {"m_1": "translation", "f": "foreign"}

        self.controller().edit_card_and_sisters(card, new_fact_data,
            new_card_type, new_tag_names=["default"], correspondence=correspondence)

        fact = self.database().fact(card.fact._id, is_id_internal=True)
        assert fact["f"] == "foreign"
        card_1, card_2 = self.database().cards_from_fact(fact)

        assert len(card_1.tags) == 1
        assert len(card_2.tags) == 1
        for tag in card_1.tags:
            assert tag.name == "default"
        for tag in card_2.tags:
            assert tag.name == "default"

        assert self.database().con.execute("select tags from cards where _id=?",
            (card_1._id, )).fetchone()[0] == "default"
        assert self.database().con.execute("select tags from cards where _id=?",
            (card_2._id, )).fetchone()[0] == "default"

        assert card_1.active == True
        assert card_2.active == True

        print(card_1.grade, card_2.grade)

        if card_1.fact_view.id == "3::my_language.1": # Recognition
            assert "foreign" in card_1.question()
            assert "translation" in card_2.question()
            assert card_1.grade == -1
            assert card_2.grade == 2
        else:
            assert "foreign" in card_2.question()
            assert "translation" in card_1.question()
            assert card_2.grade == -1
            assert card_1.grade == 2

    def test_convert_duplicates(self):
        card_type_map = self.card_type_with_id("4")
        fact_data = {"loc": "test",
                     "blank": "test1",
                     "marked": "test2"}
        card = self.controller().create_new_cards(fact_data,
            card_type_map, grade=2, tag_names=["default"])[0]

        card_type = self.card_type_with_id("1")
        fact_data = {"f": "test1",
                     "b": "test2"}
        card = self.controller().create_new_cards(fact_data,
            card_type, grade=2, tag_names=["default"])[0]

        correspondence = {"f": "blank", "b": "marked"}

        new_fact_data = {"loc": "test",
                     "blank": "test1",
                     "marked": "test2"}

        self.controller().edit_card_and_sisters(card, new_fact_data,
            card_type_map, new_tag_names=["default"], correspondence=correspondence)

    def teardown(self):
        if os.path.exists("a.ogg"):
            os.remove("a.ogg")
        MnemosyneTest.teardown(self)



#
# html_statistics.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.statistics_page import HtmlStatisticsPage


# The statistics page.

class MyHtmlStatistics(HtmlStatisticsPage):

    name = "My html staticsics"
         
    def prepare_statistics(self, variant):
        card = self.review_controller().card
        self.html = """<html<body>
        <style type="text/css">
        table { height: 100%;
                margin-left: auto; margin-right: auto;
                text-align: center}
        body  { background-color: white;
                margin: 0;
                padding: 0;
                border: thin solid #8F8F8F; }
        </style></head><table><tr><td>"""  
        self.html += "There are lies, damn lies and statistics."
        self.html += "</td></tr></table></body></html>"


# Wrap it into a Plugin and then register the Plugin.

class MyHtmlStatisticsPlugin(Plugin):
    name = "Html statistics example"
    description = "Example plugin for html statistics"
    components = [MyHtmlStatistics]
    supported_API_level = 3

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(MyHtmlStatisticsPlugin)



#
# test_logging.py <Peter.Bienstman@UGent.be>
#

import os
import shutil
import time

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne import Mnemosyne
from openSM2sync.log_entry import EventTypes
from mnemosyne.libmnemosyne.ui_components.main_widget import MainWidget


class MyMainWidget(MainWidget):

    def show_question(self, question, b, c, d):
        if question == "Delete this card?":
            return 1 # Yes
        else:
            raise NotImplementedError


class TestLogging(MnemosyneTest):

    def restart(self):
        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)

    def test_logging(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "1", "b": "b"}
        card = self.controller().create_new_cards(fact_data, card_type,
                     grade=-1, tag_names=["default"])[0]
        card_id_1 = card.id
        self.review_controller().show_new_question()
        self.review_controller().grade_answer(0)
        self.review_controller().show_new_question()
        self.review_controller().grade_answer(1)
        self.review_controller().grade_answer(4)

        self.mnemosyne.finalise()
        self.restart()
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "2", "b": "b"}
        card = self.controller().create_new_cards(fact_data, card_type,
                     grade=-1, tag_names=["default"])[0]
        self.review_controller().show_new_question()
        self.controller().delete_current_card()

        self.log().dump_to_science_log()

        sql_res = self.database().con.execute(\
            "select * from log where _id=1").fetchone()
        assert sql_res[1] == EventTypes.STARTED_PROGRAM

        sql_res = self.database().con.execute(\
            "select * from log where _id=2").fetchone()
        assert sql_res[1] == EventTypes.STARTED_SCHEDULER

        sql_res = self.database().con.execute(\
            "select * from log where _id=3").fetchone()
        assert sql_res[1] == EventTypes.LOADED_DATABASE
        assert sql_res[6] == 0
        assert sql_res[7] == 0
        assert sql_res[8] == 0

        sql_res = self.database().con.execute(\
            "select * from log where _id=11").fetchone()
        assert sql_res[1] == EventTypes.ADDED_TAG
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=12").fetchone()
        assert sql_res[1] == EventTypes.EDITED_CRITERION
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=13").fetchone()
        assert sql_res[1] == EventTypes.ADDED_FACT
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=14").fetchone()
        assert sql_res[1] == EventTypes.ADDED_CARD
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=15").fetchone()
        assert sql_res[1] == EventTypes.REPETITION
        assert sql_res[6] == 1
        assert sql_res[7] == 0
        assert sql_res[11] == 0
        assert sql_res[12] == 0
        assert sql_res[14] - sql_res[2] == 0
        assert sql_res[13] == 0
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=16").fetchone()
        assert sql_res[1] == EventTypes.REPETITION
        assert sql_res[6] == 2
        assert sql_res[7] == 0
        assert sql_res[11] == 0
        assert sql_res[12] <= 10 # Depends on CPU load.
        assert sql_res[14] - sql_res[2] == 0
        assert sql_res[13] == 0

        sql_res = self.database().con.execute(\
            "select * from log where _id=17").fetchone()
        assert sql_res[1] == EventTypes.REPETITION
        assert sql_res[6] == 3
        assert sql_res[7] == 0
        assert sql_res[11] == 0
        assert sql_res[12] <= 10 # Depends on CPU load.
        new_interval = sql_res[14] - sql_res[2]
        assert new_interval > 0
        assert sql_res[13] == 0

        sql_res = self.database().con.execute(\
            "select * from log where _id=18").fetchone()
        assert sql_res[1] == EventTypes.SAVED_DATABASE
        assert sql_res[6] == 0
        assert sql_res[7] == 0
        assert sql_res[8] == 1

        sql_res = self.database().con.execute(\
            "select * from log where _id=19").fetchone()
        assert sql_res[1] == EventTypes.STOPPED_PROGRAM

        sql_res = self.database().con.execute(\
            "select * from log where _id=20").fetchone()
        assert sql_res[1] == EventTypes.STARTED_PROGRAM

        sql_res = self.database().con.execute(\
            "select * from log where _id=21").fetchone()
        assert sql_res[1] == EventTypes.STARTED_SCHEDULER

        sql_res = self.database().con.execute(\
            "select * from log where _id=22").fetchone()
        assert sql_res[1] == EventTypes.LOADED_DATABASE
        assert sql_res[6] == 0
        assert sql_res[7] == 0
        assert sql_res[8] == 1

        sql_res = self.database().con.execute(\
            "select * from log where _id=30").fetchone()
        assert sql_res[1] == EventTypes.ADDED_FACT
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=31").fetchone()
        assert sql_res[1] == EventTypes.ADDED_CARD

        sql_res = self.database().con.execute(\
            "select * from log where _id=32").fetchone()
        assert sql_res[1] == EventTypes.DELETED_CARD
        assert sql_res[3] is not None

        sql_res = self.database().con.execute(\
            "select * from log where _id=33").fetchone()
        assert sql_res[1] == EventTypes.DELETED_FACT
        assert sql_res[3] is not None

        self.config()["upload_science_logs"] = True
        self.database().dump_to_science_log()

        logfile = os.path.join(os.path.abspath("dot_test"), "log.txt")
        found = False
        for line in open(logfile):
            if "R " + card_id_1 + " 4" in line:
                found = True
                assert str(new_interval) + " 0 | 0.0" in line
        assert found == True

    def test_unique_index(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_2 = self.card_type_with_id("2")
        card_1, card_2 = self.controller().create_new_cards(fact_data, card_type_2,
                                              grade=-1, tag_names=["default"])
        log_index = self.database().con.execute(\
            """select _id from log order by _id desc limit 1""").fetchone()[0]
        # Note: we need to keep the last log entry intact, otherwise indexes
        # start again at 1 and mess up the sync.
        self.database().con.execute("""delete from log where _id <?""", (log_index,))
        self.database().save()
        self.database().con.execute("""vacuum""")
        fact_data = {"f": "question2",
                     "b": "answer2"}
        card_type_2 = self.card_type_with_id("1")
        card_1  = self.controller().create_new_cards(fact_data, card_type_2,
                                              grade=-1, tag_names=["default"])
        assert self.database().con.execute(\
            """select _id from log order by _id limit 1""").fetchone()[0] \
            == log_index

    def test_recover_user_id(self):
        assert self.config()["user_id"] is not None
        MnemosyneTest.teardown(self)

        open(os.path.join(os.getcwd(), "dot_test", "history", "userid_001.bz2"), "w")
        os.remove(os.path.join(os.getcwd(), "dot_test", "config.db"))

        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()

        assert self.config()["user_id"] == "userid"

    def test_recover_user_id_2(self):
        assert self.config()["user_id"] is not None
        MnemosyneTest.teardown(self)

        open(os.path.join(os.getcwd(), "dot_test", "history", "userid_machine_001.bz2"), "w")
        os.remove(os.path.join(os.getcwd(), "dot_test", "config.db"))

        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()

        assert self.config()["user_id"] == "userid"

    def test_log_index_of_last_upload_1(self):
        assert self.log().log_index_of_last_upload() == 0

    def test_log_index_of_last_upload_2(self):
        machine_id = self.config().machine_id()
        for filename in ["user_001.bz2", "user_%s_2.bz2" % machine_id]:
            open(os.path.join(os.getcwd(), "dot_test", "history", filename), "w")
        assert self.log().log_index_of_last_upload() == 2

    def test_log_index_of_last_upload_3(self):
        machine_id = self.config().machine_id()
        for filename in ["user_001.bz2"]:
            open(os.path.join(os.getcwd(), "dot_test", "history", filename), "w")
        assert self.log().log_index_of_last_upload() == 1

    def test_log_index_of_last_upload_4(self):
        machine_id = self.config().machine_id()
        for filename in ["user_005.bz2"]:
            open(os.path.join(os.getcwd(), "dot_test", "history", filename), "w")
        assert self.log().log_index_of_last_upload() == 5

    def test_log_index_of_last_upload_5(self):
        machine_id = self.config().machine_id()
        for filename in ["user_othermachine_005.bz2"]:
            open(os.path.join(os.getcwd(), "dot_test", "history", filename), "w")
        assert self.log().log_index_of_last_upload() == 0

    def test_log_index_of_last_upload_6(self):
        machine_id = self.config().machine_id()
        for filename in ["user_othermachine_005.bz2", "user_%s_2.bz2" % machine_id]:
            open(os.path.join(os.getcwd(), "dot_test", "history", filename), "w")
        assert self.log().log_index_of_last_upload() == 2

    def test_log_index_of_last_upload_7(self):
        machine_id = self.config().machine_id()
        for filename in ["user_001.bz2", "user_othermachine_005.bz2", "user_%s_2.bz2" % machine_id]:
            open(os.path.join(os.getcwd(), "dot_test", "history", filename), "w")
        assert self.log().log_index_of_last_upload() == 2

    def test_log_upload(self):
        machine_id_file = os.path.join(self.mnemosyne.config().config_dir, "machine.id")
        f = open(machine_id_file, "w")
        print("TESTMACHINE", file=f)
        f.close()
        self.config().change_user_id("UPLOADTEST")
        self.config()["max_log_size_before_upload"] = 1
        MnemosyneTest.teardown(self)

        self.mnemosyne = Mnemosyne(upload_science_logs=True, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()
        MnemosyneTest.teardown(self)

        self.mnemosyne = Mnemosyne(upload_science_logs=True, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()
        MnemosyneTest.teardown(self)

        self.mnemosyne = Mnemosyne(upload_science_logs=True, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()


    def test_log_upload_bad_server(self):
        # Most reliable way of setting this variable is throug config.py, otherwise
        # it will stay alive in a dangling imported userconfig.
        config_py_file = os.path.join(self.mnemosyne.config().config_dir, "config.py")
        f = open(config_py_file, "w")
        print("science_server = \"noserver:80\"", file=f)
        f.close()

        machine_id_file = os.path.join(self.mnemosyne.config().config_dir, "machine.id")
        f = open(machine_id_file, "w")
        print("TESTMACHINE", file=f)
        f.close()
        self.config().change_user_id("UPLOADTEST")
        self.config()["max_log_size_before_upload"] = 1
        MnemosyneTest.teardown(self)

        self.mnemosyne = Mnemosyne(upload_science_logs=True, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()
        MnemosyneTest.teardown(self)

        self.mnemosyne = Mnemosyne(upload_science_logs=True, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("test_logging", "MyMainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.mnemosyne.start_review()

    def mem_importer(self):
        for format in self.mnemosyne.component_manager.all("file_format"):
            if format.__class__.__name__ == "Mnemosyne1Mem":
                return format

    def test_archive_old_logs(self):
        # Import old history.
        filename = os.path.join(os.getcwd(), "tests", "files", "basedir_bz2",
                                "default.mem")
        self.mem_importer().do_import(filename)
        assert self.database().con.execute("select count() from log").fetchone()[0] == 23
        assert not os.path.exists(os.path.join("dot_test", "archive"))
        # Archive.
        self.database().archive_old_logs()
        assert self.database().con.execute("select count() from log").fetchone()[0] == 12
        archive_name = os.listdir(os.path.join(os.getcwd(), "dot_test", "archive"))[0]
        archive_path = os.path.join(os.getcwd(), "dot_test", "archive", archive_name)
        import sqlite3
        arch_con = sqlite3.connect(archive_path)
        assert arch_con.execute("select count() from log").fetchone()[0] == 11

    def test_log_warn_about_too_many_cards(self):
        timestamp = int(time.time())
        self.database().log_warn_about_too_many_cards(timestamp)
        results = self.database().con.execute("""select timestamp from log WHERE event_type=? and timestamp = ?""", (EventTypes.WARNED_TOO_MANY_CARDS, timestamp)).fetchall()
        assert 1 == len(results)



#
# language.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class Language(Component):

    """Interface class describing the functions to be implemented by the
    actual language classes.

    Contains all language-specific calls related to parsing, stemming,
    dictionary lookup, text-to-speech, ... .

    The link between languages and card types is stored in config.db, so that
    it easily fits with the syncing logic and also works for built-in uncloned
    card types.

    """

    component_type = "language"
    name = None
    used_for = None  # ISO 639-1 code
    sublanguages = {} # {"en_GB": "English (UK)", "en_US": "English (US")}.
    feature_description = None



#
# mnemosyne_format.py <Peter.Bienstman@UGent.be>
#

import os
import sqlite3
import tempfile

from openSM2sync.log_entry import EventTypes
from mnemosyne.libmnemosyne.utils import copy


class MnemosyneFormat(object):

    def __init__(self, database):
        self.database = database

    def supports(self, program_name, program_version, database_version):
        return program_name.lower() == "mnemosyne" and \
            database_version == self.database.version

    def binary_filename(self, store_pregenerated_data, interested_in_old_reps):
        self.database.release_connection()
        # Copy the database to a temporary file.
        self.tmp_name = os.path.join(os.path.dirname(self.database._path),
            "__FORSTREAMING__.db")
        copy(self.database._path, self.tmp_name)
        # Delete old reps if needed.
        if not interested_in_old_reps:
            con = sqlite3.connect(self.tmp_name)
            con.execute("delete from log where event_type=?",
                (EventTypes.REPETITION, ))
            con.commit()
            con.execute("vacuum")
            con.close()
        # Delete pregerated data if needed.
        if not store_pregenerated_data:
            con = sqlite3.connect(self.tmp_name)
            con.executescript("""
            begin;
            drop index i_cards;
            create table cards_new(
                _id integer primary key,
                id text,
                card_type_id text,
                _fact_id integer,
                fact_view_id text,
                grade integer,
                next_rep integer,
                last_rep integer,
                easiness real,
                acq_reps integer,
                ret_reps integer,
                lapses integer,
                acq_reps_since_lapse integer,
                ret_reps_since_lapse integer,
                creation_time integer,
                modification_time integer,
                extra_data text default "",
                scheduler_data integer default 0,
                active boolean default 1
            );
            insert into cards_new select _id, id, card_type_id, _fact_id,
                fact_view_id, grade, next_rep, last_rep, easiness, acq_reps,
                ret_reps, lapses, acq_reps_since_lapse, ret_reps_since_lapse,
                creation_time, modification_time, extra_data, scheduler_data,
                active from cards;
            drop table cards;
            alter table cards_new rename to cards;
            create index i_cards on cards (id);
            commit;
            vacuum;
            """)
            con.close()
        return self.tmp_name

    def clean_up(self):
        os.remove(self.tmp_name)



#
# qt_translator.py <Johannes.Baiter@gmail.com>
#

import os
import sys

from PyQt5.QtCore import QTranslator, QCoreApplication, QLocale

from mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator \
     import GetTextGuiTranslator


class QtGuiTranslator(GetTextGuiTranslator):

    """This deals with translating Qt itself. The rest of the strings are
    still translated using the gettext mechanism, as we've modified pyuic4
    to use gettext too.

    """

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.qt_translator = QTranslator(QCoreApplication.instance())
        try:
            self.qt_dir = os.environ["QTDIR"]
        except:
            if sys.platform == "win32":
                self.qt_dir = os.path.join(sys.exec_prefix, "share", "qt5")
            else:
                self.qt_dir = os.path.join("/usr", "share", "qt5")
        # Avoid stuff like Thai numerals if the language is not explicitly
        # set to Thai.
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))

    def translate_ui(self, language):
        app = QCoreApplication.instance()
        # We always need to remove a translator, to make sure we generate a
        # LanguageChange event even if their is no Qt translation for that
        # language installed.
        app.removeTranslator(self.qt_translator)
        self.qt_translator.load(os.path.join(self.qt_dir, "translations",
            "qt_" + language + ".qm"))
        app.installTranslator(self.qt_translator)
        # The title can also contain the database name.
        self.controller().update_title()




#
# english.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class English(Language):

    name = _("English")
    used_for = "en"
    sublanguages = {"en-GB": _("English (UK)"),
                     "en-US": _("English (US)"),
                     "en-IN": _("English (India)"),
                     "en-AU": _("English (Australian)")}
    feature_description = _("Google translation and text-to-speech.")



#
# component.py <Peter.Bienstman@UGent.be>
#


class Component(object):

    """Base class of components that are registered with the component
    manager. This is a list of component types: config, log, database,
    scheduler, stopwatch, gui_translator, card_type, card_type_converter,
    render_chain, renderer, filter, card_type_widget,
    generic_card_type_widget, ui_component, controller, main_widget,
    review_controller, review_widget, file format, plugin, hook,
    criterion, criterion_applier, statistics_page, sync_server,
    study_mode, translator, pronouncer, all the abstract dialogs, ...

    'used_for' can store certain relationships between components, e.g.
    a card type widget is used for a certain card type.

    Most of the time, instances are stored here, apart from widgets in which
    case classes are stored. (Instantiating a complex widget can take a lot of
    time on a mobile device, and should be done lazily.) Only the main
    widget is stored as an instance here.

    To achieve this lazy instantiation for widget, set 'instantiate == LATER'.
    Other components can then instantiate the widget when they see fit.
    The instance is not cached for subsequent reuse, as these widgets
    typically can become obsolete/overwritten by plugins.

    It can be that when instantiating a component, not all the other components
    on which it relies have been instantiated yet. E.g., the log and the
    database depend on each other before doing actual work. Therefore, some of
    the  actual initialisation work can be postponed to the 'activate' function.

    Each component has access to all of the context of the other components
    because it hold a reference to the user's component manager.

    We need to pass the context of the component manager already in the
    constructor, as many component make use of it in their __init__ method.
    This means that derived components should always call the
    Component.__init__ if they provide their own constructor.

    In case the GUI needs to add functionality to a certain component, that
    can be done through component_manager.add_gui_to_component() function
    (lower level) or through the gui_for_component class variable (higher
    level).

    """

    component_type = ""
    used_for = None

    IMMEDIATELY = 0
    LATER = 1

    instantiate = IMMEDIATELY

    def __init__(self, component_manager, **kwds):
        super().__init__(**kwds)  # For parent classes other than 'Object'.
        self.component_manager = component_manager
        self.gui_components = []
        self.instantiated_gui_components = []

    def activate(self):

        """Initialisation code called when the component is about to do actual
        work, and which can't happen in the constructor, e.g. because
        components on which it relies have not yet been registered.

        """

        pass

    def activate_gui_components(self):

        """GUI classes are only instantiated when activated, since that can take
        a lot of time on mobile clients.

        Classes that require more control over this, e.g. when needing to return
        values from the UI, can subclass this.

        """

        for component in self.gui_components:
            component = component(component_manager=self.component_manager)
            self.component_manager.register(component)
            component.activate()
            self.instantiated_gui_components.append(component)

    def deactivate(self):
        for component in self.instantiated_gui_components:
            component.deactivate()
            self.component_manager.unregister(component)
        self.instantiated_gui_components = []

    # Convenience functions, for easier access to all of the context of
    # libmnemosyne from within a component.

    def _(self):
        return self.component_manager.current("gui_translator")

    def gui_translator(self):
        return self.component_manager.current("gui_translator")

    def config(self):
        return self.component_manager.current("config")

    def log(self):
        return self.component_manager.current("log")

    def database(self):
        return self.component_manager.current("database")

    def scheduler(self):
        return self.component_manager.current("scheduler")

    def stopwatch(self):
        return self.component_manager.current("stopwatch")

    def main_widget(self):
        return self.component_manager.current("main_widget")

    def review_widget(self):
        return self.component_manager.current("review_widget")

    def controller(self):
        return self.component_manager.current("controller")

    def review_controller(self):
        return self.component_manager.current("review_controller")

    def card_types(self):
        return self.component_manager.all("card_type")

    def card_type_with_id(self, id):
        return self.component_manager.card_type_with_id[id]

    def render_chain(self, id="default"):
        return self.component_manager.render_chain_with_id[id]

    def study_mode_with_id(self, id):
        return self.component_manager.study_mode_with_id[id]

    def plugins(self):
        return self.component_manager.all("plugin")

    def languages(self):
        return self.component_manager.language_with_id.values()

    def language_with_id(self, id):
        return self.component_manager.language_with_id[id]

    def flush_sync_server(self):

        """If there are still dangling sessions (i.e. those waiting in vain
        for more client input) in the sync server, we should flush them and
        make sure they restore from backup before doing anything that could
        change the database (e.g. adding a card). Otherwise, if these
        sessions close later during program shutdown, their backup
        restoration will override the changes.

        Also stop any running media.

        """

        server = self.component_manager.current("sync_server")
        if server:
            server.flush()
        review_widget = self.review_widget()
        if review_widget:
            review_widget.stop_media()



#
# card_type_tree_wdgt.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.libmnemosyne.criteria.default_criterion import DefaultCriterion

# Code reuse through inheritance.
from mnemosyne.pyqt_ui.tag_tree_wdgt import TagDelegate, TagsTreeWdgt

# We hijack QTreeWidgetItem a bit and store extra data in a hidden column, so
# that we don't need to implement a custom tree model.
# The first column stores the string displayed, i.e. the card type id.
# For card types, the second column stores the node id, in this case
# the card type id.

DISPLAY_STRING = 0
NODE = 1

class CardTypeDelegate(TagDelegate):

    def setEditorData(self, editor, index):
        # Get rid of the card count.
        self.previous_node_name = \
            index.model().data(index).rsplit(" (", 1)[0]
        node_index = index.model().index(index.row(), NODE, index.parent())
        self.card_type_id = index.model().data(node_index)
        editor.setText(self.previous_node_name)

    def commit_and_close_editor(self):
        editor = self.sender()
        if self.previous_node_name == editor.text():
            self.redraw_node.emit(self.card_type_id)
        else:
            self.rename_node.emit(self.card_type_id, editor.text())
        self.closeEditor.emit(editor, QtWidgets.QAbstractItemDelegate.NoHint)


class CardTypesTreeWdgt(TagsTreeWdgt):

    """Displays all the card types in a tree together with check boxes."""

    card_types_changed_signal = QtCore.pyqtSignal()

    def __init__(self, acquire_database=None, **kwds):
        super().__init__(**kwds)
        self.delegate = CardTypeDelegate(\
            component_manager=kwds["component_manager"], parent=self)
        self.tree_wdgt.setItemDelegate(self.delegate)
        self.delegate.rename_node.connect(self.rename_node)
        self.delegate.redraw_node.connect(self.redraw_node)
        self.acquire_database = acquire_database

    def menu_rename(self):
        nodes = self.selected_nodes_which_can_be_renamed()
        # If there are tags selected, this means that we could only have got
        # after pressing return on an actual edit, due to our custom
        # 'keyPressEvent'. We should not continue in that case.
        if len(nodes) == 0:
            return

        from mnemosyne.pyqt_ui.ui_rename_card_type_dlg \
            import Ui_RenameCardTypeDlg

        class RenameDlg(QtWidgets.QDialog, Ui_RenameCardTypeDlg):
            def __init__(self, old_card_type_name):
                super().__init__()
                self.setupUi(self)
                self.card_type_name.setText(old_card_type_name)

        old_card_type_name = self.card_type_with_id(nodes[0]).name
        dlg = RenameDlg(old_card_type_name)
        if dlg.exec_() == QtWidgets.QDialog.Accepted:
            self.rename_node(nodes[0], dlg.card_type_name.text())

    def menu_delete(self):
        nodes = self.selected_nodes_which_can_be_deleted()
        if len(nodes) == 0:
            return
        if len(nodes) > 1:
            question = _("Delete these card types?")
        else:
            question = _("Delete this card type?")
        answer = self.main_widget().show_question\
            (question, _("&OK"), _("&Cancel"), "")
        if answer == 1: # Cancel.
            return
        self.delete_nodes(nodes)

    def display(self, criterion=None):
        # Create criterion if needed.
        if criterion is None:
            criterion = DefaultCriterion(self.component_manager)
        # Determine number of cards at each level of the tree.
        root_count = 0
        count_for_card_type = {}
        count_for_fact_view = {}
        for card_type in self.card_types():
            card_type_count = 0
            for fact_view in card_type.fact_views:
                count = self.database().card_count_for_fact_view \
                    (fact_view, active_only=False)
                card_type_count += count
                count_for_fact_view[fact_view] = count
            count_for_card_type[card_type] = card_type_count
            root_count += card_type_count
        # Fill widget.
        self.nodes_which_can_be_deleted = []
        self.nodes_which_can_be_renamed = []
        self.tree_wdgt.clear()
        self.node_items = []
        self.card_type_fact_view_ids_for_node_item = []
        root_item = QtWidgets.QTreeWidgetItem(self.tree_wdgt,
            [_("All card types (%d)") % root_count], 0)
        root_item.setFlags(root_item.flags() | \
           QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsTristate)
        root_item.setCheckState(0, QtCore.Qt.Checked)
        for card_type in self.database().sorted_card_types():
            if card_type.hidden_from_UI:
                continue
            card_type_item = QtWidgets.QTreeWidgetItem(root_item, ["%s (%d)" % \
                (_(card_type.name), count_for_card_type[card_type])], 0)
            card_type_item.setFlags(card_type_item.flags() | \
                QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsTristate)
            card_type_item.setCheckState(0, QtCore.Qt.Checked)
            card_type_item.setData(NODE, QtCore.Qt.DisplayRole,
                    QtCore.QVariant(card_type.id))
            if count_for_card_type[card_type] == 0 and \
                self.database().is_user_card_type(card_type):
                    self.nodes_which_can_be_deleted.append(card_type.id)
            if self.database().is_user_card_type(card_type):
                self.nodes_which_can_be_renamed.append(card_type.id)
                card_type_item.setFlags(card_type_item.flags() | \
                    QtCore.Qt.ItemIsEditable)
            for fact_view in card_type.fact_views:
                fact_view_item = QtWidgets.QTreeWidgetItem(card_type_item,
                    ["%s (%d)" % (_(fact_view.name),
                    count_for_fact_view[fact_view])], 0)
                fact_view_item.setFlags(fact_view_item.flags() | \
                    QtCore.Qt.ItemIsUserCheckable)
                if (card_type.id, fact_view.id) in \
                    criterion.deactivated_card_type_fact_view_ids:
                    check_state = QtCore.Qt.Unchecked
                else:
                    check_state = QtCore.Qt.Checked
                fact_view_item.setCheckState(0, check_state)
                # Since fact_view_item seems mutable, we cannot use a dict.
                self.node_items.append(fact_view_item)
                self.card_type_fact_view_ids_for_node_item.append(\
                    (card_type.id, fact_view.id))
        self.tree_wdgt.expandAll()

    def checked_to_criterion(self, criterion):
        criterion.deactivated_card_type_fact_view_ids = set()
        for i in range(len(self.node_items)):
            item = self.node_items[i]
            card_type_fact_view_ids = \
                self.card_type_fact_view_ids_for_node_item[i]
            if item.checkState(0) == QtCore.Qt.Unchecked:
                criterion.deactivated_card_type_fact_view_ids.add(\
                    card_type_fact_view_ids)
        return criterion

    def save_criterion(self):
        self.saved_criterion = DefaultCriterion(self.component_manager)
        self.checked_to_criterion(self.saved_criterion)
        # Now we've saved the checked state of the tree.
        # Saving and restoring the selected state is less trivial, because
        # in the case of trees, the model indexes have parents which become
        # invalid when creating the widget.
        # The solution would be to save tags and reselect those in the new
        # widget.

    def restore_criterion(self):
        # See if any card types have been deleted meanwhile.
        old_card_type_ids = set([cursor[0] for cursor in \
            self.saved_criterion.deactivated_card_type_fact_view_ids])
        deleted_card_type_ids = set()
        for card_type_id in old_card_type_ids:
            try:
                card_type = self.card_type_with_id(card_type_id)
            except:
                deleted_card_type_ids.add(card_type_id)
        # Remove offending entries.
        deactivated_card_type_fact_view_ids = \
            self.saved_criterion.deactivated_card_type_fact_view_ids
        for card_type_id, fact_view_id in deactivated_card_type_fact_view_ids:
            if card_type_id in deleted_card_type_ids:
                self.saved_criterion.deactivated_card_type_fact_view_ids.\
                    discard((card_type_id, fact_view_id))
        self.display(self.saved_criterion)

    def rename_node(self, node, new_name):
        if self.acquire_database:
            self.acquire_database()
        self.save_criterion()
        card_type = self.card_type_with_id(node)
        self.controller().rename_card_type(card_type, new_name)
        self.restore_criterion()
        self.card_types_changed_signal.emit()

    def delete_nodes(self, nodes):
        if self.acquire_database:
            self.acquire_database()
        self.save_criterion()
        for node in nodes:
            card_type = self.card_type_with_id(node)
            self.controller().delete_card_type(card_type)
        self.restore_criterion()
        self.card_types_changed_signal.emit()

    def rebuild(self):

        """To be called when external events invalidate the tag tree,
        e.g. due to edits in the card browser widget.

        """

        self.save_criterion()
        # Now we've saved the checked state of the tree.
        # Saving and restoring the selected state is less trivial, because
        # in the case of trees, the model indexes have parents which become
        # invalid when creating the widget.
        # The solution would be to save card types and reselect those in the
        # new widget.
        self.restore_criterion()



#
# qtextedit2.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _


class QTextEdit2(QtWidgets.QTextEdit):

    """QTextEdit with extra options in popup menu."""

    def __init__(self, parent, card_type, pronunciation_hiding):

        """'pronunciation_hiding' is set to None when there is no
        pronunciation key to hide, and to True or False when there is one to
        hide.

        """

        super().__init__(parent)
        self.pronunciation_hiding = pronunciation_hiding
        self.card_type = card_type
        # Make list of translators and pronouncers for this card type.
        language_id = self.card_type.config().card_type_property(\
            "language_id", self.card_type, default=None)
        self.translators = [] if not language_id else \
            self.card_type.component_manager.all("translator", language_id)
        self.pronouncers = [] if not language_id else \
            self.card_type.component_manager.all("pronouncer", language_id)

    def contextMenuEvent(self, e):
        popup = self.createStandardContextMenu()
        popup.addSeparator()
        popup.addAction(_("Insert &image"), self.insert_img,
                        QtGui.QKeySequence(_("Ctrl+I")))
        popup.addAction(_("Insert &sound"), self.insert_sound,
                        QtGui.QKeySequence(_("Ctrl+S")))
        popup.addAction(_("Insert vi&deo"), self.insert_video,
                        QtGui.QKeySequence(_("Ctrl+D")))
        popup.addAction(_("Insert &Flash"), self.insert_flash,
                        QtGui.QKeySequence(_("Ctrl+F")))
        # Pronunciation hiding.
        if self.pronunciation_hiding in [True, False]:
            popup.addSeparator()
            self.hide_action = QtWidgets.QAction(\
                _("&Hide pronunciation field for this card type"), popup)
            self.hide_action.setCheckable(True)
            self.hide_action.setChecked(self.pronunciation_hiding)
            self.hide_action.toggled.connect(\
                self.parent().pronunciation_hiding_toggled)
            popup.addAction(self.hide_action)
        # Translators.
        if len(self.translators):
            popup.addSeparator()
            for translator in self.translators:
                translator_action = QtWidgets.QAction(\
                    translator.popup_menu_text, popup)
                translator_action.triggered.connect(\
                    lambda checked, t=translator: self.translate(t))
                translator_action.setShortcuts([QtGui.QKeySequence("Ctrl+T")])
                popup.addAction(translator_action)
        # Pronouncers.
        if len(self.pronouncers):
            popup.addSeparator()
            for pronouncer in self.pronouncers:
                pronouncer_action = QtWidgets.QAction(\
                    pronouncer.popup_menu_text, popup)
                pronouncer_action.triggered.connect(\
                    lambda checked, p=pronouncer: self.pronounce(p))
                pronouncer_action.setShortcuts([QtGui.QKeySequence("Ctrl+P")])
                popup.addAction(pronouncer_action)
        # Show popup.
        popup.exec_(e.globalPos())

    def translate(self, translator):
        foreign_text = self.parent().foreign_text()
        translated_text = translator.show_dialog(self.card_type, foreign_text)
        if translated_text:
            self.insertPlainText(translated_text)

    def pronounce(self, pronouncer):
        foreign_text = self.parent().foreign_text()
        pronounced_text = pronouncer.show_dialog(self.card_type, foreign_text)
        if pronounced_text:
            self.insertPlainText(pronounced_text)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_I and event.modifiers() == \
            QtCore.Qt.ControlModifier:
            self.insert_img()
        elif event.key() == QtCore.Qt.Key_S and event.modifiers() == \
            QtCore.Qt.ControlModifier:
            self.insert_sound()
        elif event.key() == QtCore.Qt.Key_D and event.modifiers() == \
            QtCore.Qt.ControlModifier:
            self.insert_video()
        elif event.key() == QtCore.Qt.Key_F and event.modifiers() == \
            QtCore.Qt.ControlModifier:
            self.insert_flash()
        elif event.key() == QtCore.Qt.Key_T and event.modifiers() == \
            QtCore.Qt.ControlModifier and self.translators:
            self.translate(self.translators[-1])
        elif event.key() == QtCore.Qt.Key_P and event.modifiers() == \
            QtCore.Qt.ControlModifier and self.pronouncers:
            self.translate(self.pronouncers[-1])
        else:
            QtWidgets.QTextEdit.keyPressEvent(self, event)

    def insert_img(self):
        filter = "(*.png *.gif *.jpg *.bmp *.jpeg *.svg *.tiff *.webp" + \
                 " *.PNG *.GIF *.JPG *.BMP *.JPEG *.SVG *.TIFF *.WEBP)"
        filename = self.parent().controller().show_insert_img_dialog(filter)
        if filename:
            self.insertPlainText("<img src=\"" + filename + "\">")

    def insert_sound(self):
        filter = "(*.wav *.mp3 *.ogg *.WAV *.MP3 *.OGG)"
        filename = self.parent().controller().show_insert_sound_dialog(filter)
        if filename:
            self.insertPlainText("<audio src=\"" + filename + "\">")

    def insert_video(self):
        filter = "(*.mov *.ogg *.ogv *.mp4 *.qt" + \
                 " *.MOV *.OGG *.OGV *.MP4 *.QT)"
        filename = self.parent().controller().show_insert_video_dialog(filter)
        if filename:
            self.insertPlainText("<video src=\"" + filename + "\">")

    def insert_flash(self):
        filter = "(*.swf *.SWF)"
        filename = self.parent().controller().show_insert_flash_dialog(filter)
        if filename:
            self.insertPlainText(\
                "<object type=\"application/x-shockwave-flash\" data=\"" \
                + filename + "\" />")



#
# review_wdgt.py <Peter.Bienstman@UGent.be>
#

import os
from string import Template

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.ui_components.review_widget import ReviewWidget


class ReviewWdgt(ReviewWidget):

    """I've tried fiddling with css to get the grades area always show up at
    the bottom of the screen, no matter the contents of the cards, but I
    never got this to work satisfactory both on Firefox and IE. There would
    always be issues with spurious scrollbars or the card areas not expanding
    to fill the entire screen.
    Therefore, we place the grades at the top, where they are also always at
    the same location for easy ergonomic access.

    """

    def __init__(self, component_manager):
        ReviewWidget.__init__(self, component_manager)
        self.review_controller().set_render_chain("web_server")
        self.client_on_same_machine_as_server = False
        self.question_label = ""
        self.question = ""
        self.is_question_box_visible = True
        self.answer = ""
        self.answer_label = _("Answer:")
        self.is_answer_box_visible = True
        self.show_button = ""
        self.is_show_button_enabled = True
        self.is_grade_buttons_enabled = False
        self.status_bar = ""
        self.template = Template("""
<!DOCTYPE html>
<html>
<head>
  <title>Mnemosyne</title>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
<style type="text/css">
table {
  width: 100%;
  border: 1px solid black;
  padding: 1em;
}
table.buttonarea {
  border: 0;
  padding: 0;
  background-color: white;
}
input {
  width: 100%;
}

$card_css
</style>
</head>
<body>

<table class="buttonarea">
  <tr>
  $buttons
  </tr>
</table>

<p>$question_label</p>

<table id="mnem1" class="mnem">
  <tr>
    <td>$question</td>
  </tr>
</table>

<p>$answer_label</p>

<table id="mnem1" class="mnem">
  <tr>
    <td>$answer</td>
  </tr>
</table>

<p>$status_bar</p>

</body>
</html>
""")

    def set_client_on_same_machine_as_server\
        (self, client_on_same_machine_as_server):
        self.client_on_same_machine_as_server = \
            client_on_same_machine_as_server

    def redraw_now(self):
        pass

    def show_answer(self):
        self.review_controller().show_answer()

    def grade_answer(self, grade):
        self.review_controller().grade_answer(grade)

    def set_question_box_visible(self, is_visible):
        self.is_question_box_visible = is_visible

    def set_answer_box_visible(self, is_visible):
        self.is_answer_box_visible = is_visible

    def set_question_label(self, text):
        self.question_label = text

    def set_question(self, text):
        self.question = text

    def set_answer(self, text):
        self.answer = text

    def clear_question(self):
        self.question = ""

    def clear_answer(self):
        self.answer = ""

    def update_show_button(self, text, is_default, is_enabled):
        self.show_button = text
        self.is_show_button_enabled = is_enabled

    def set_grades_enabled(self, is_enabled):
        self.is_grade_buttons_enabled = is_enabled

    def set_default_grade(self, grade):
        pass

    def update_status_bar_counters(self):
        scheduled_count, non_memorised_count, active_count = \
            self.review_controller().counters()
        counters = "Sch.: %d Not mem.: %d Act.: %d" % \
            (scheduled_count, non_memorised_count, active_count)
        self.status_bar = """
              <table class="buttonarea">
              <tr>
                <td> """ + counters + """ </td>
                <td>
                  <form action="" method="post">
                    <input type="submit" name="star" value="Star">
                  </form>
                </td>
              </tr></table>"""

    def to_html(self):
        self.controller().heartbeat()
        card_css = ""
        card = self.review_controller().card
        if card:
            card_css = self.render_chain().renderer_for_card_type\
                (card.card_type).card_type_css(card.card_type)
        buttons = ""
        if self.is_grade_buttons_enabled:
            buttons = ""
            for i in range(6):
                buttons += """
                  <td>
                    <form action="" method="post">
                      <input type="submit" name="grade" accesskey="%d"
                       value="%d">
                    </form>
                  </td>""" % (i, i)
        if self.is_show_button_enabled:
            buttons = """
              <td>
                <form action="" method="post">
                  <input type="submit" name="show_answer" value="%s">
                </form>
              </td>""" % (self.show_button)
        if not self.question:
            self.question = "&nbsp;"  # For esthetic reasons.
        if not self.answer:
            self.answer = "&nbsp;"
        extended_status_bar = self.status_bar
        if self.client_on_same_machine_as_server:
            extended_status_bar = extended_status_bar.replace(\
            "</tr></table>", \
            """<td>
                  <form action="" method="post">
                    <input type="submit" name="exit" value="Exit">
                  </form>
                </td>
            </tr></table>""")
        return self.template.substitute(card_css=card_css, buttons=buttons,
            question_label=self.question_label, question=self.question,
            answer_label=self.answer_label, answer=self.answer,
            status_bar=extended_status_bar).encode("utf-8")




#
# new_card_type_widget.py <Peter.Bienstman@UGent.be>
#

# Colour the widget for the front to back card type red.

from PyQt5 import QtGui

from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.card_types.front_to_back import FrontToBack
from mnemosyne.pyqt_ui.card_type_wdgt_generic import GenericCardTypeWdgt


# Don't forget to inherit from CardTypeWidget located in
# mnemosyne.libmnemosyne.ui_components.card_type_widget!
# (Here this is already fullfilled because GenericCardTypeWdgt itself inherits
# from CardTypeWidget.)
# We inherit from GenericCardTypeWdgt purely for implementation reasons.
# What we are building here is a CardTypeWidget (specific for the FrontToBack
# card type), not a GenericCardTypeWidget, so we need to respecify the
# component_type.

class RedCardTypeWdgt(GenericCardTypeWdgt):

    component_type = "card_type_widget"
    used_for = FrontToBack

    def __init__(self, **kwds):
        super().__init__(**kwds)
        for edit_box in self.edit_boxes:
            p = QtGui.QPalette()
            p.setColor(QtGui.QPalette.Active, QtGui.QPalette.Base, \
                       QtGui.QColor("red"))
            edit_box.setPalette(p)


# Wrap it into a Plugin and then register the Plugin.

class RedPlugin(Plugin):
    
    name = "Red"
    description = "Red widget for front-to-back cards"
    components = [RedCardTypeWdgt]
    supported_API_level = 3

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(RedPlugin)







#
# html5_video.py <Peter.Bienstman@UGent.be>
#

import re

from mnemosyne.libmnemosyne.filter import Filter

re_video = re.compile(r"""<video( src=\".+?\")>""", re.DOTALL | re.IGNORECASE)


class Html5Video(Filter):

    """Add autoplay and control tags to html5 video tags."""

    def run(self, text, card, fact_key, **render_args):
        options = ""
        if self.config()["media_autoplay"]:
            options += " autoplay=1"
        if self.config()["media_controls"]:
            options += " controls=1"
        return re.sub(re_video, r"<video\1" + options + ">", text)





#
# activate_cards_dlg.py <Peter.Bienstman@UGent.be>
#

import _dialogs

from mnemosyne.libmnemosyne.ui_components.dialogs import ActivateCardsDialog


class ActivateCardsDlg(ActivateCardsDialog):

    def activate(self):
        ActivateCardsDialog.activate(self)
        self.criteria_by_name = {}
        active_set_name = ""
        active_criterion = self.database().current_criterion()
        for criterion in self.database().criteria():
            if criterion._id != 1:
                self.criteria_by_name[criterion.name] = criterion
                if criterion == active_criterion:
                    active_set_name = criterion.name
        _dialogs.activate_cards_dlg_activate(\
            "____".join(sorted(self.criteria_by_name.keys())), active_set_name)



#
# stopwatch.py <Peter.Bienstman@UGent.be>
#

import time

from mnemosyne.libmnemosyne.component import Component


class Stopwatch(Component):

    """The main use of the stop watch is to measure the time it takes the user
    to answer a question. 'start_time' is used to keep track of when the card
    was first shown.

    """

    component_type = "stopwatch"

    def __init__(self, component_manager):
        Component.__init__(self, component_manager)
        self.start_time = 0
        self.start_time_of_leg = 0
        self.running_time = 0

    def start(self):
        self.start_time = time.time()
        self.start_time_of_leg = self.start_time
        self.running_time = 0

    def pause(self):
        self.running_time += time.time() - self.start_time_of_leg

    def unpause(self):
        self.start_time_of_leg = time.time()

    def stop(self):
        # Don't reset start_time here yet, as we rely on it to know when we
        # first showed the card.
        self.running_time += time.time() - self.start_time_of_leg
        return self.running_time

    def time(self):
        return self.running_time




#
# somali.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Somali(Language):

    name = _("Somali")
    used_for = "so"
    feature_description = _("Google translation.")



#
# database_maintenance.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.database import DatabaseMaintenance


class AndroidDatabaseMaintenance(DatabaseMaintenance):

    def run(self):
        # Use shown_question here, since this is implemented to block.
        answer = self.main_widget().show_question(\
_("About to archive old logs to improve running speed. Depending on the size of your database and the speed of your device, this can take 10 minutes or more. Please leave Mnemosyne running in the foreground."),
        _("OK, proceed"), "", "")
        if answer == 0:
            DatabaseMaintenance.run(self)




#
# test_non_latin_font_size_increase.py <Peter.Bienstman@UGent.be>
#

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne.filters.escape_to_html import EscapeToHtml

    
class TestNonLatinFontSizeIncrease(MnemosyneTest):

    def test_1(self):
        fact_data = {"f": "question " + chr(40960),
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                              grade=-1, tag_names=[])[0]

        self.config()["non_latin_font_size_increase"] = 2
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" in card.question()
        
    def test_2(self):
        fact_data = {"f": "<protect>question " + chr(40960) + "</protect>",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                              grade=-1, tag_names=[])[0]

        self.config()["non_latin_font_size_increase"] = 2
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" not in card.question()

    def test_3(self):
        fact_data = {"f": "question " + chr(40960),
                     "p_1": "",
                     "m_1": "answer"}
        card_type = self.card_type_with_id("3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                              grade=-1, tag_names=[])[0]
        self.config()["non_latin_font_size_increase"] = 2
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" not in card.question()
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" not in card.answer()
        
    def test_4(self):
        fact_data = {"f": "question " + chr(40960),
                     "m_1": "answer"}
        card_type = self.card_type_with_id("3")
        self.controller().clone_card_type(card_type, "my_3")
        card_type = self.card_type_with_id("3::my_3")
        card = self.controller().create_new_cards(fact_data, card_type,
                                              grade=-1, tag_names=[])[0]

        self.config()["non_latin_font_size_increase"] = 2
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" not in card.question()
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" not in card.answer()
        
    def test_5(self):
        fact_data = {"f": "question " + chr(40960) + "tail",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
                                              grade=-1, tag_names=[])[0]

        self.config()["non_latin_font_size_increase"] = 2
        assert """<font style=\"font-size:14pt\">""" + chr(40960) + "</font>" in card.question()



#
# front_to_back.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.libmnemosyne.fact_view import FactView


class FrontToBack(CardType):

    id = "1"
    name = _("Front-to-back only")

    # List and name the keys.
    fact_keys_and_names = [("f", _("Front")),
                           ("b", _("Back"))]

    # Front-to-back.
    v = FactView(_("Front-to-back"), "1.1")
    v.q_fact_keys = ["f"]
    v.a_fact_keys = ["b"]

    fact_views = [v]
    unique_fact_keys = ["f"]
    required_fact_keys = ["f"]



#
# text_format.py <Peter.Bienstman@UGent.be>
#

              
class TextFormat(object):
    
    """Class handling the conversion from log entries and other data to
    text streams and vice versa.

    """

    mime_type = None

    def repr_partner_info(self, info):
        raise NotImplementedError

    def parse_partner_info(self, text):
        raise NotImplementedError

    def log_entries_header(self, number_of_entries):
        raise NotImplementedError

    def log_entries_footer(self):
        raise NotImplementedError

    def repr_log_entry(self, log_entry):
        raise NotImplementedError

    def parse_log_entries(self, txt):
        raise NotImplementedError

    def repr_message(self, message, traceback=None):
        raise NotImplementedError

    def parse_message(self, text):
        raise NotImplementedError



#
# malayalam.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Malayalam(Language):

    name = _("Malayalam")
    used_for = "ml"
    feature_description = _("Google translation.")



#
#  anki2.py Peter.Bienstman@UGent.be>
#

import os
import re
import json
import time
import shutil
import sqlite3
import zipfile

from mnemosyne.libmnemosyne.fact import Fact
from mnemosyne.libmnemosyne.card import Card
from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.fact_view import FactView
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.libmnemosyne.file_format import FileFormat
from mnemosyne.libmnemosyne.card_types.M_sided import MSided
from mnemosyne.libmnemosyne.criteria.default_criterion import \
     DefaultCriterion
from mnemosyne.libmnemosyne.file_formats.media_preprocessor \
    import MediaPreprocessor

sound_re = re.compile("(?i)(\[sound:(?P<fname>[^]]+)\])")


class TagCache(dict, Component):

    def __init__(self, component_manager):
        Component.__init__(self, component_manager)
        self._tags_with_name = {}

    def __getitem__(self, tag_name):
        if not tag_name in self._tags_with_name:
            self._tags_with_name[tag_name] = \
                self.database().get_or_create_tag_with_name(tag_name)
        return self._tags_with_name[tag_name]


class Anki2(FileFormat, MediaPreprocessor):

    description = _("Anki2 collections")
    extension = ".anki2"
    filename_filter = _("Anki2 collections (*.anki2 *.apkg)")
    import_possible = True
    export_possible = False

    def __init__(self, component_manager):
        FileFormat.__init__(self, component_manager)
        MediaPreprocessor.__init__(self, component_manager)

    def extract_apkg(self, filename):
        # Extract zipfile.
        w = self.main_widget()
        w.set_progress_text(_("Decompressing..."))
        zip_file = zipfile.ZipFile(filename, "r")
        tmp_dir = os.path.join(os.path.dirname(filename), "__TMP__")
        if os.path.exists(tmp_dir):
            shutil.rmtree(tmp_dir)
        zip_file.extractall(tmp_dir)
        # Rename media.
        media_file_with_id = json.loads(\
            open(os.path.join(tmp_dir, "media")).read())
        number_of_files = len(media_file_with_id)
        w.set_progress_text(_("Processing media files..."))
        w.set_progress_range(number_of_files)
        w.set_progress_update_interval(number_of_files/50)
        media_dir = os.path.join(tmp_dir, "collection.media")
        os.mkdir(media_dir)
        for id, media_file in media_file_with_id.items():
            w.increase_progress(1)
            if media_file.startswith("latex-"):
                continue
            os.rename(os.path.join(tmp_dir, id),
                      os.path.join(media_dir, media_file))
        # Note: Anki itself goes through some trouble of making sure ids
        # are unique and updating them if needed, because their id scheme
        # is very sensitive to collisions. We decided not to do that here
        # because it would break being able to reimport an updated version
        # of the cards.
        return tmp_dir

    def do_import(self, filename, extra_tag_names=""):
        self.main_widget().show_information(_(\
"Note that while you can edit imported cards, adding new cards to Anki's card types is currently not supported.\n\nAlso, in case you run into problems, don't hesitate to contact the developers."))
        FileFormat.do_import(self, filename, extra_tag_names)
        w = self.main_widget()
        db = self.database()
        # Preprocess apkg files.
        tmp_dir = None
        if filename.endswith(".apkg"):
            tmp_dir = self.extract_apkg(filename)
            filename = os.path.join(tmp_dir, "collection.anki2" )
        # Set up tag cache.
        tag_with_name = TagCache(self.component_manager)
        # Open database.
        con = sqlite3.connect(filename)
        # Copy media directory.
        w.set_progress_text(_("Copying media files..."))
        src = filename.replace(".anki2", ".media")
        dst = db.media_dir()
        number_of_files = len(os.listdir(src))
        w.set_progress_range(number_of_files)
        w.set_progress_update_interval(number_of_files/50)
        for item in os.listdir(src):
            shutil.copy(os.path.join(src, item), os.path.join(dst, item))
            w.increase_progress(1)
        # Import collection table.
        w.set_progress_text(_("Importing card types..."))
        # Too few in number to warrant counted progress bar.
        card_type_for_mid = {}  # mid: model id
        deck_name_for_did = {}  # did: deck id
        for id, crt, mod, scm, ver, dty, usn, ls, conf, models, decks, \
            dconf, tags in con.execute("""select id, crt, mod, scm, ver, dty,
            usn, ls, conf, models, decks, dconf, tags from col"""):
            # mod: modification time, ignore.
            # scm: schema modification time, ignore.
            # ver: schema version, ignore.
            # dty: no longer used according to Anki source.
            # usn: syncing related, ignore.
            # ls: last sync, ignore.
            # conf: configuration, ignore.
            # dconf: deck configuration, ignore.
            # tags: list of tags, but they turn up later in the notes, ignore.
            collection_creation_time = crt
            decks = json.loads(decks)
            # Decks will be converted to Tags when creating cards.
            for did in decks:
                deck_name_for_did[int(did)] = decks[did]["name"]
            # Models will be converted to CardTypes
            models = json.loads(models)
            for mid in models:  # mid: model id
                card_type_id = "7::" + mid
                card_type_already_imported = \
                    db.has_card_type_with_id(card_type_id)
                if card_type_already_imported:
                    card_type = self.component_manager.card_type_with_id[\
                        card_type_id]
                else:
                    card_type = MSided(self.component_manager)
                card_type.name = models[mid]["name"]
                card_type.id = card_type_id
                card_type.hidden_from_UI = False
                card_type_for_mid[int(mid)] = card_type
                #vers = models[mid]["vers"] # Optional version, ignore.
                #tags = models[mid]["tags"] # Seems empty, ignore.
                #did = models[mid]["did"] # Deck id, ignore.
                #usn = models[mid]["usn"] # Syncing related, ignore.
                if "req" in models[mid]:
                    required = models[mid]["req"]
                    # Cache for a calculation to determine which fields are
                    # required. "req": [[0, "all", [0]]]
                    # Not yet implemented.
                else:
                    required = []
                flds = models[mid]["flds"]
                flds.sort(key=lambda x : x["ord"])
                card_type.fact_keys_and_names = []
                for field in flds:
                    card_type.fact_keys_and_names.append(\
                        (str(field["ord"]), field["name"]))
                    #media = field["media"] # Reserved for future use, ignore.
                    #sticky = field["sticky"] # Sticky field, ignore.
                    #rtl = field["rtl"] # Text direction, ignore.
                    font_string = field["font"] + "," + str(field["size"]) + \
                        ",-1,5,50,0,0,0,0,0,Regular"
                    self.config().set_card_type_property("font", font_string,
                        card_type, str(field["ord"]))
                #sortf = models[mid]["sortf"] # Sorting field, ignore.
                tmpls = models[mid]["tmpls"]
                tmpls.sort(key=lambda x : x["ord"])
                # Fact views.
                card_type.fact_views = []
                for template in tmpls:
                    fact_view_id = card_type.id + "." + str(template["ord"])
                    fact_view_already_imported = \
                        db.has_fact_view_with_id(fact_view_id)
                    if fact_view_already_imported:
                        fact_view = db.fact_view(\
                            fact_view_id, is_id_internal=False)
                        fact_view.name = template["name"]
                    else:
                        fact_view = FactView(template["name"], fact_view_id)
                    fact_view.extra_data["qfmt"] = template["qfmt"]
                    fact_view.extra_data["afmt"] = template["afmt"]
                    fact_view.extra_data["bqfmt"] = template["bqfmt"]
                    fact_view.extra_data["bafmt"] = template["bafmt"]
                    fact_view.extra_data["ord"] = template["ord"]
                    #did = template["did"] # Deck id, ignore.
                    card_type.fact_views.append(fact_view)
                    if fact_view_already_imported:
                        db.update_fact_view(fact_view)
                    else:
                        db.add_fact_view(fact_view)
                #mod = models[mid]["mod"] # Modification time, ignore.
                type_ = models[mid]["type"] # 0: standard, 1 cloze
                id = models[mid]["id"]
                css = models[mid]["css"]
                #latex_preamble = models[mid]["latexPre"] # Ignore.
                #latex_postamble = models[mid]["latexPost"] # Ignore.
                # Save to database.
                card_type.extra_data = {"css":css, "id":id, "type":type_}
                if card_type_already_imported:
                    db.update_card_type(card_type)
                else:
                    db.add_card_type(card_type)
        # nid are Anki-internal indices for notes, so we need to temporarily
        # store some information.
        tag_names_for_nid = {}
        card_type_for_nid = {}
        # Import facts and tags.
        w.set_progress_text(_("Importing notes..."))
        number_of_notes = con.execute("select count() from notes").fetchone()[0]
        w.set_progress_range(number_of_notes)
        w.set_progress_update_interval(number_of_notes/20)
        fact_for_nid = {}
        modification_time_for_nid = {}
        for id, guid, mid, mod, usn, tags, flds, sfld, csum, flags, data in \
            con.execute("""select id, guid, mid, mod, usn, tags, flds, sfld,
            csum, flags, data from notes"""):
            # usn: syncing related, ignore.
            # sfld: sorting field, ignore.
            # csum: checksum, ignore.
            # flags: seems empty, ignore.
            # data: seems empty, ignore.
            # Make compatible with openSM2sync:
            guid = guid.replace("`", "ap").replace("\"", "qu")
            guid = guid.replace("&", "am").replace("<", "lt").replace(">", "gt")
            modification_time_for_nid[id] = mod
            card_type = card_type_for_mid[int(mid)]
            card_type_for_nid[id] = card_type
            fields = flds.split("\x1f")
            assert (len(fields) == len(card_type.fact_keys_and_names))
            fact_data = {}
            for i in range(len(fields)):
                fact_key = card_type.fact_keys_and_names[i][0]
                data = fields[i]
                # Deal with sound tags.
                for match in sound_re.finditer(data):
                    fname = match.group("fname")
                    data = data.replace(\
                        match.group(0), "<audio src=\"" + fname + "\">")
                # Deal with latex tags.
                data = data.replace("[latex]", "<latex>")
                data = data.replace("[/latex]", "</latex>")
                data = data.replace("[$]", "<$>")
                data = data.replace("[/$]", "</$>")
                data = data.replace("[$$]", "<$$>")
                data = data.replace("[/$$]", "</$$>")
                fact_data[fact_key] = data
            if db.has_fact_with_id(guid):
                fact = db.fact(guid, is_id_internal=False)
                fact.data = fact_data
                db.update_fact(fact)
            else:
                fact = Fact(fact_data, id=guid)
                db.add_fact(fact)
            fact_for_nid[id] = fact
            tag_names_for_nid[id] = tags
            w.increase_progress(1)
        # Import logs. This needs to happen before creating the cards,
        # otherwise, the sync protocol will use the scheduling data from the
        # latest repetition log, instead of the correct current one.
        w.set_progress_text(_("Importing logs..."))
        number_of_logs = con.execute("select count() from revlog").fetchone()[0]
        w.set_progress_range(number_of_logs)
        w.set_progress_update_interval(number_of_logs/20)
        for id, cid, usn, ease, ivl, lastIvl, factor, time, type_ in \
            con.execute("""select id, cid, usn, ease, ivl, lastIvl, factor,
            time, type from revlog"""):
            # usn: syncing related, ignore.
            if type_ == 0:  # Acquisition phase.
                grade = 0
            else:  # Retention phase.
                grade = ease + 1  # Anki ease is from 1 to 4.
            timestamp = int(id/1000)
            scheduled_interval = lastIvl*86400 if lastIvl > 0 else 0
            new_interval = ivl*86400 if ivl > 0 else 0
            next_rep = timestamp + new_interval
            easiness = factor/1000 if factor else 2.5
            db.log_repetition(timestamp=timestamp, card_id=cid, grade=grade,
                easiness=easiness, acq_reps=0, ret_reps=0, lapses=0,
                acq_reps_since_lapse=0, ret_reps_since_lapse=0,
                scheduled_interval=scheduled_interval,
                actual_interval=scheduled_interval,
                thinking_time=int(time/1000), next_rep=next_rep,
                scheduler_data=0)
            w.increase_progress(1)
        # Import cards.
        w.set_progress_text(_("Importing cards..."))
        number_of_cards = con.execute("select count() from cards").fetchone()[0]
        w.set_progress_range(number_of_cards)
        w.set_progress_update_interval(number_of_cards/20)
        for id, nid, did, ord, mod, usn, type_, queue, due, ivl, factor, reps, \
            lapses, left, odue, odid, flags, data in con.execute("""select id,
            nid, did, ord, mod, usn, type, queue, due, ivl, factor, reps,
            lapses, left, odue, odid, flags, data from cards"""):
            # type: 0=new, 1=learning, 2=due
            # queue: same as above, and -1=suspended,
            #        -2=user buried, -3=sched buried
            # due is used differently for different queues.
            # - new queue: note id or random int
            # - rev queue: integer day
            # - lrn queue: integer timestamp
            # In Mnemosyne, type=2 / rev queue corresponds to grades >= 2.
            # mod: modification time, but gets updated on each answer.
            # usn: syncing related, ignore.
            # left: repetitions left to graduation, ignore.
            # odue: original due, related to filtered decks, ignore.
            # odid: original deck id, related to filtered decks, ignore.
            # flags: seems empty, ignore.
            # data: seems empty, ignore
            fact = fact_for_nid[nid]
            card_type = card_type_for_nid[nid]
            creation_time = int(nid / 1000)
            if card_type.extra_data["type"] == 0:
                fact_view = card_type.fact_views[ord]
            else:  # Cloze.
                fact_view = card_type.fact_views[0]
            already_imported = db.has_card_with_id(id)
            if already_imported:
                card = db.card(id, is_id_internal=False)
                card.card_type = card_type
                card.fact = fact
                card.fact_view = fact_view
                card.creation_time = creation_time
            else:
                card = Card(card_type, fact, fact_view,
                            creation_time=creation_time)
            card.id = id
            card.extra_data["ord"] = ord  # Needed separately for clozes.
            tag_names = [tag_name.strip() for \
                             tag_name in extra_tag_names.split(",")]
            tag_names += [tag_name.strip() for \
                             tag_name in tag_names_for_nid[nid].split(" ")]
            tag_names += [deck_name_for_did[did].strip().replace(",", ";")]
            for tag_name in tag_names:
                if tag_name:
                    card.tags.add(tag_with_name[tag_name])
            card.next_rep = collection_creation_time + due*86400
            card.last_rep = card.next_rep - ivl*86400
            card.easiness = factor/1000 if factor else 2.5
            card.acq_reps = 1   # No information.
            card.ret_reps = reps
            card.lapses = lapses
            card.acq_reps_since_lapse = card.acq_reps  # No information.
            card.ret_reps_since_lapse = card.ret_reps  # No information.
            card.modification_time = modification_time_for_nid[nid]
            self.active = (queue >= 0)
            if type_ == 0:  # 'new', unseen.
                card.reset_learning_data()
            elif type_ == 1:  # 'learning', acquisition phase.
                card.grade = 0
                card.last_rep = mod
                card.next_rep = mod
            else:  # 'due', retention phase.
                card.grade = 4  # No information.
            if card.grade >= 2:
                assert card.ret_reps_since_lapse != 0 # Issue #93 on github.
            if already_imported:
                db.update_card(card)
            else:
                db.add_card(card)
            w.increase_progress(1)
        # Create criteria for 'database' tags.
        for deck_name in deck_name_for_did.values():
            deck_name = deck_name.strip().replace(",", ";")
            if deck_name in [criterion.name for criterion in db.criteria()]:
                continue
            tag = tag_with_name[deck_name]
            criterion = DefaultCriterion(\
                component_manager=self.component_manager)
            criterion.name = deck_name
            criterion._tag_ids_active.add(tag._id)
            criterion._tag_ids_forbidden = set()
            db.add_criterion(criterion)
        # Clean up.
        con.close()
        if tmp_dir:
            shutil.rmtree(tmp_dir)
        w.close_progress()
        self.warned_about_missing_media = False



#
# activate_cards_dlg.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.pyqt_ui.card_set_name_dlg import CardSetNameDlg
from mnemosyne.pyqt_ui.ui_activate_cards_dlg import Ui_ActivateCardsDlg
from mnemosyne.libmnemosyne.ui_components.dialogs import ActivateCardsDialog
from mnemosyne.pyqt_ui.tip_after_starting_n_times import \
     TipAfterStartingNTimes

class ActivateCardsDlg(QtWidgets.QDialog, ActivateCardsDialog,
                       TipAfterStartingNTimes, Ui_ActivateCardsDlg):

    started_n_times_counter = "started_activate_cards_n_times"
    tip_after_n_times = \
        {3 : _("If you find yourself selecting the same tags and card types many types, you can press the button 'Save this set for later use' to give it a name to select it more quickly later."),
         6 : _("Double-click on the name of a saved set to quickly activate it and close the dialog."),
         9 : _("You can right-click on the name of a saved set to rename or delete it."),
         12 : _("If you single-click the name of a saved set, modifications to the selected tags and card types are not saved to that set unless you press 'Save this set for later use' again. This allows you to make some quick-and-dirty temporary modifications.")}

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        # Initialise widgets.
        self.was_showing_a_saved_set = False
        self.is_shutting_down = False
        criterion = self.database().current_criterion()
        self.criterion_classes = \
            self.component_manager.all("criterion")
        current_criterion = self.database().current_criterion()
        self.widget_for_criterion_type = {}
        for criterion_class in self.criterion_classes:
            widget = self.component_manager.current\
                ("criterion_widget", used_for=criterion_class)\
                (component_manager=self.component_manager, parent=self)
            self.tab_widget.addTab(widget, criterion_class.criterion_type)
            self.widget_for_criterion_type[criterion_class.criterion_type] \
                = widget
        self.tab_widget.setCurrentWidget(self.widget_for_criterion_type\
                                         [current_criterion.criterion_type])
        self.tab_widget.tabBar().setVisible(self.tab_widget.count() > 1)
        self.tab_widget.currentWidget().display_criterion(current_criterion)
        # Restore state.
        state = self.config()["activate_cards_dlg_state"]
        if state:
            self.restoreGeometry(state)
        splitter_state = self.config()["activate_cards_dlg_splitter_state"]
        if not splitter_state:
            self.splitter.setSizes([100, 350])
        else:
            self.splitter.restoreState(splitter_state)
        # Should go last, otherwise the selection of the saved sets pane will
        # always be cleared.
        self.update_saved_sets_pane()

    def keyPressEvent(self, event):
        if event.key() in [QtCore.Qt.Key_Delete, QtCore.Qt.Key_Backspace]:
            self.delete_set()
        else:
            QtWidgets.QDialog.keyPressEvent(self, event)

    def change_widget(self, index):
        self.saved_sets.clearSelection()

    def activate(self):
        ActivateCardsDialog.activate(self)
        self.exec_()

    def update_saved_sets_pane(self):
        self.saved_sets.clear()
        self.criteria_by_name = {}
        active_name = ""
        active_criterion = self.database().current_criterion()
        for criterion in self.database().criteria():
            if criterion._id != 1:
                self.criteria_by_name[criterion.name] = criterion
                self.saved_sets.addItem(criterion.name)
                if criterion == active_criterion:
                    active_name = criterion.name
        self.saved_sets.sortItems()
        if active_name:
            item = self.saved_sets.findItems(active_name,
                QtCore.Qt.MatchExactly)[0]
            self.saved_sets.setCurrentItem(item)
            self.was_showing_a_saved_set = True
        else:
            self.saved_sets.clearSelection()
            self.was_showing_a_saved_set = False
        splitter_sizes = self.splitter.sizes()
        if self.saved_sets.count() == 0:
            self.splitter.setSizes([0, sum(splitter_sizes)])
        else:
            if splitter_sizes[0] == 0: # First time we add a set.
                self.splitter.setSizes([0.3* sum(splitter_sizes),
                    0.7 * sum(splitter_sizes)])

    def saved_sets_custom_menu(self, pos):
        menu = QtWidgets.QMenu()
        menu.addAction(_("Delete"), self.delete_set)
        menu.addAction(_("Rename"), self.rename_set)
        menu.exec_(self.saved_sets.mapToGlobal(pos))

    def save_set(self):
        criterion = self.tab_widget.currentWidget().criterion()
        if criterion.is_empty():
            self.main_widget().show_error(\
                _("This set can never contain any cards!"))
            return
        CardSetNameDlg(criterion, self.criteria_by_name.keys(),
                       component_manager=self.component_manager, parent=self).exec_()
        if not criterion.name:  # User cancelled.
            return
        if criterion.name in self.criteria_by_name.keys():
            answer = self.main_widget().show_question(_("Update this set?"),
                _("&OK"), _("&Cancel"), "")
            if answer == 1:  # Cancel.
                return
            original_criterion = self.criteria_by_name[criterion.name]
            criterion._id = original_criterion._id
            criterion.id = original_criterion.id
            self.database().update_criterion(criterion)
        else:
            self.database().add_criterion(criterion)
        self.update_saved_sets_pane()
        item = self.saved_sets.findItems(criterion.name,
            QtCore.Qt.MatchExactly)[0]
        self.saved_sets.setCurrentItem(item)
        if self.config()["showed_help_on_renaming_sets"] == False:
            self.main_widget().show_information(\
                _("You can right-click on the name of a saved set to rename or delete it."))
            self.config()["showed_help_on_renaming_sets"] = True

    def delete_set(self):
        if not self.saved_sets.currentItem():
            return
        answer = self.main_widget().show_question(_("Delete this set?"),
            _("&OK"), _("&Cancel"), "")
        if answer == 1:  # Cancel.
            return -1
        else:
            name = self.saved_sets.currentItem().text()
            criterion = self.criteria_by_name[name]
            self.database().delete_criterion(criterion)
            self.database().save()
            self.update_saved_sets_pane()

    def rename_set(self):
        name = self.saved_sets.currentItem().text()
        criterion = self.criteria_by_name[name]
        criterion.name = name
        other_names = list(self.criteria_by_name.keys())
        other_names.remove(name)
        CardSetNameDlg(criterion, other_names,
                       component_manager=self.component_manager,
                       parent=self).exec_()
        if criterion.name == name:  # User cancelled.
            return
        self.database().update_criterion(criterion)
        self.database().save()
        self.update_saved_sets_pane()
        item = self.saved_sets.findItems(criterion.name,
            QtCore.Qt.MatchExactly)[0]
        self.saved_sets.setCurrentItem(item)

        # load_set gets triggered by ItemActivated, but this does not happen
        # when the user changes the sets through the arrow keys (Qt bug?).
        # Therefore, we also catch currentItemChanged and forward it to
        # change_set, but to prevent unwanted firing when loading the widget
        # for the first time (which would erase the current criterion in case
        # it is not a saved criterion), we discard this event if previous_item
        # is None.
        #
        # To test when editing this code: initial start, with and without
        # current criterion being a saved criterion, changing the set through
        # clicking or through the arrows.

    def load_set(self, item, dummy=None):
        # Sometimes item is None, e.g. during the process of deleting a saved
        # set, so we need to discard the event then.
        if item is None:
            return
        name = item.text()
        criterion = self.criteria_by_name[name]
        self.tab_widget.setCurrentWidget(self.widget_for_criterion_type\
                                             [criterion.criterion_type])
        self.tab_widget.currentWidget().display_criterion(criterion)
        # Restore the selection that got cleared in change_widget.
        item = self.saved_sets.findItems(criterion.name,
            QtCore.Qt.MatchExactly)[0]
        self.saved_sets.setCurrentItem(item)
        self.was_showing_a_saved_set = True

    def change_set(self, item, previous_item):
        if previous_item is not None:
            self.load_set(item)

    def select_set_and_close(self, item):
        self.load_set(item)
        # Work around a Qt bug where these calls would still fire when clicking
        # in the same area where e.g. the tag browser used to be, even after
        # closing the 'Activate cards' window.
        self.is_shutting_down = True
        self.accept()

    def _store_state(self):
        self.config()["activate_cards_dlg_state"] = \
            self.saveGeometry()
        self.config()["activate_cards_dlg_splitter_state"] = \
            self.splitter.saveState()

    def closeEvent(self, event):
        # Generated when clicking the window's close button.
        self._store_state()
        # This allows the state of the tag tree to be saved.
        self.tab_widget.currentWidget().close()

    def accept(self):
        criterion = self.tab_widget.currentWidget().criterion()
        if criterion.is_empty():
            self.main_widget().show_error(\
                _("This set can never contain any cards!"))
            return
        if self.saved_sets.count() != 0 and self.config()\
            ["showed_help_on_double_clicking_sets"] == False:
            self.main_widget().show_information(\
_("You can double-click on the name of a saved set to activate it and close the dialog."))
            self.config()["showed_help_on_double_clicking_sets"] = True
        if len(self.saved_sets.selectedItems()) > 0:
            criterion.name = self.saved_sets.currentItem().text()
        self.database().set_current_criterion(criterion)
        # 'accept' does not generate a close event.
        self._store_state()
        return QtWidgets.QDialog.accept(self)



#
# manage_card_types_dlg.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.pyqt_ui.card_type_language_list_wdgt import \
     CardTypeLanguageListWdgt
from mnemosyne.pyqt_ui.clone_card_type_dlg import CloneCardTypeDlg
from mnemosyne.pyqt_ui.ui_manage_card_types_dlg import Ui_ManageCardTypesDlg
from mnemosyne.pyqt_ui.edit_M_sided_card_type_dlg import EditMSidedCardTypeDlg
from mnemosyne.libmnemosyne.ui_components.dialogs import ManageCardTypesDialog


class ManageCardTypesDlg(QtWidgets.QDialog, ManageCardTypesDialog,
                         Ui_ManageCardTypesDlg):

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.native_card_types = CardTypeLanguageListWdgt(\
            parent=self.native_card_types_box,
            component_manager=self.component_manager)
        self.vertical_layout_native_card_types.insertWidget(\
            0, self.native_card_types)
        self.Anki_card_types = CardTypeLanguageListWdgt(\
            parent=self.Anki_card_types_box,
            component_manager=self.component_manager)
        self.vertical_layout_Anki_card_types.insertWidget(\
            0, self.Anki_card_types)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        self.update()
        state = self.config()["manage_card_types_dlg_state"]
        if state:
            self.restoreGeometry(state)

    def activate(self):
        ManageCardTypesDialog.activate(self)
        self.exec_()

    def update(self):
        # Fill up native types panel.
        card_types = []
        for card_type in self.database().sorted_card_types():
            if not card_type.hidden_from_UI and \
               not card_type.id.startswith("7"):
                card_types.append(card_type)
        self.native_card_types.set_card_types(card_types)
        self.rename_native_card_type_button.setEnabled(False)
        self.delete_native_card_type_button.setEnabled(False)
        self.native_card_types.selectionModel().currentRowChanged.connect(\
            self.activate_native_card_type)
        # Fill up Anki card types panel.
        card_types = []
        for card_type in self.database().sorted_card_types():
            if self.database().is_user_card_type(card_type) and \
               not card_type.hidden_from_UI and \
               card_type.id.startswith("7"):
                card_types.append(card_type)
        self.Anki_card_types.set_card_types(card_types)
        self.edit_Anki_card_type_button.setEnabled(False)
        self.rename_Anki_card_type_button.setEnabled(False)
        self.delete_Anki_card_type_button.setEnabled(False)
        self.Anki_card_types.selectionModel().currentRowChanged.connect(\
            self.activate_Anki_card_type)
        if len(card_types) == 0:
            self.Anki_card_types_box.hide()
        else:
            self.native_card_types_box.setTitle(_("Mnemosyne card types"))

    def clone_card_type(self):
        if not self.config()["clone_help_shown"]:
            self.main_widget().show_information(\
_("Here, you can make clones of existing card types. This allows you to format cards in this type independently from cards in the original type. E.g. you can make a clone of 'Vocabulary', call it 'Thai' and set a Thai font specifically for this card type without disturbing your other cards."))
            self.config()["clone_help_shown"] = True
        dlg = CloneCardTypeDlg(parent=self, component_manager=self.component_manager)
        dlg.exec_()
        self.update()

    def activate_native_card_type(self):
        self.rename_native_card_type_button.setEnabled(True)
        self.delete_native_card_type_button.setEnabled(True)
        self.edit_Anki_card_type_button.setEnabled(False)
        self.rename_Anki_card_type_button.setEnabled(False)
        self.delete_Anki_card_type_button.setEnabled(False)

    def activate_Anki_card_type(self):
        self.rename_native_card_type_button.setEnabled(False)
        self.delete_native_card_type_button.setEnabled(False)
        self.edit_Anki_card_type_button.setEnabled(True)
        self.rename_Anki_card_type_button.setEnabled(True)
        self.delete_Anki_card_type_button.setEnabled(True)

    def delete_native_card_type(self):
        self.delete_selected_card_type(\
            self.native_card_types.selected_card_type())

    def delete_Anki_card_type(self):
        self.delete_selected_card_type(\
            self.Anki_card_types.selected_card_type())

    def delete_selected_card_type(self, card_type):
        if not card_type:
            return
        answer = self.main_widget().show_question\
            (_("Delete this card type?"), _("&OK"), _("&Cancel"), "")
        if answer == 1: # Cancel.
            return
        self.controller().delete_card_type(card_type)
        self.update()

    def rename_native_card_type(self):
        self.rename_selected_card_type(\
            self.native_card_types.selected_card_type())

    def rename_Anki_card_type(self):
        self.rename_selected_card_type(\
            self.Anki_card_types.selected_card_type())

    def rename_selected_card_type(self, card_type):
        if not card_type:
            return

        from mnemosyne.pyqt_ui.ui_rename_card_type_dlg \
            import Ui_RenameCardTypeDlg

        class RenameDlg(QtWidgets.QDialog, Ui_RenameCardTypeDlg):
            def __init__(self, old_card_type_name):
                super().__init__()
                self.setupUi(self)
                self.card_type_name.setText(old_card_type_name)

        dlg = RenameDlg(card_type.name)
        if dlg.exec_() == QtWidgets.QDialog.Accepted:
            new_name = dlg.card_type_name.text()
            self.controller().rename_card_type(card_type, new_name)
            self.update()

    def edit_Anki_card_type(self):
        card_type = self.Anki_card_types.selected_card_type()
        if not card_type:
            return
        dlg = EditMSidedCardTypeDlg(card_type,
            component_manager=self.component_manager)
        if dlg.exec_() == QtWidgets.QDialog.Accepted:
            self.database().update_card_type(card_type)
            self.update()

    def _store_state(self):
        self.config()["manage_card_types_dlg_state"] = self.saveGeometry()

    def closeEvent(self, event):
        # Generated when clicking the window's close button.
        self._store_state()

    def accept(self):
        self._store_state()
        return QtWidgets.QDialog.accept(self)

    def reject(self):
        self._store_state()
        return QtWidgets.QDialog.reject(self)




#
# find_and_replace.py <Peter.Bienstman@UGent.be>
#

import copy
from mnemosyne.script import Mnemosyne

# 'data_dir = None' will use the default system location, edit as appropriate.
data_dir = None
mnemosyne = Mnemosyne(data_dir)

find_string = "\xa0"
replace_string = " "

for _card_id, _fact_id in mnemosyne.database().cards():
    card = mnemosyne.database().card(_card_id, is_id_internal=True)
    changed = False
    new_fact_data = copy.copy(card.fact.data)
    for fact_key in card.fact.data:
        if find_string in card.fact[fact_key]:
            new_fact_data[fact_key] = \
                card.fact[fact_key].replace(find_string, replace_string)
            print((new_fact_data[fact_key]))
            changed = True
    if changed:
        mnemosyne.controller().edit_card_and_sisters(card, new_fact_data,
            card.card_type, [tag.name for tag in card.tags], {})
mnemosyne.finalise()



#
# add_card.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.script import Mnemosyne

# 'data_dir = None' will use the default system location, edit as appropriate.
data_dir = None
mnemosyne = Mnemosyne(data_dir)

# For info on the card types and their different fields, see
# libmnemosyne/card_types
fact_data = {"f": "front", "b": "back"}
card_type = mnemosyne.card_type_with_id("1")
mnemosyne.controller().create_new_cards(fact_data,
    card_type, grade=4, tag_names=["tag_1", "tag_2"])

mnemosyne.finalise()



#
# card_browser_render_chain.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.render_chain import RenderChain
from mnemosyne.libmnemosyne.renderers.html_css_card_browser \
     import HtmlCssCardBrowser
from mnemosyne.libmnemosyne.filters.escape_to_html_for_card_browser \
     import EscapeToHtmlForCardBrowser
from mnemosyne.libmnemosyne.filters.RTL_handler import RTLHandler


class CardBrowserRenderChain(RenderChain):

    """Renders either the question or answer for display in the card browser.

    """

    id = "card_browser"

    never_join_q_to_a = True

    filters = [EscapeToHtmlForCardBrowser, RTLHandler]
    renderers = [HtmlCssCardBrowser]




#
# hook.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class Hook(Component):

    """Function hooks are used by registering an instance of this class as
    component of type hook, with the 'used_for' argument any of the following
    hook points:

       =====================================   ===============================
       'after_load'                            in database.load
       'after_backup'                          in database.backup
       'before_unload'                         in database.unload
       'configuration_defaults'                in configuration.set_defaults
       'before_repetititon'                    in SM2_mnemosyne.grade_answer
                                               extra argument: card
       'after_repetititon'                     in SM2_mnemosyne.grade_answer
                                               extra argument: card
       'dynamically_create_media_files'        in SQLite_sync
                                               extra argument: data
       'delete_unused_media_files'             in SQLite_sync
       'preprocess_cloze'                      in cloze.py
       'postprocess_q_a_cloze'                 in cloze.py
       'at_rollover'                           in default_controller.py
       =====================================   ===============================

    It is the 'run' method that will get called at the corresponding point
    in the program.

    """

    component_type = "hook"

    def run(self):
        raise NotImplementedError



#
# simple_html5_audio.py <Peter.Bienstman@UGent.be>
#

import re
import urllib.request, urllib.parse, urllib.error

from mnemosyne.libmnemosyne.filter import Filter

re_audio = re.compile(r"""<audio src=\"(.+?)\"(.*?)>""",
    re.DOTALL | re.IGNORECASE)


class SimpleHtml5Audio(Filter):

    """Most simple html5 audio player to ensure maximum compatibility across
    a wide range of browsers and mobile devices.

    Issues:

     - no autoplay to prevent synchronisation bugs.
     - not very well suited for multiple audio files.
     - no support for start and stop tags.

    """

    def run(self, text, card, fact_key, **render_args):
        if not re_audio.search(text):
            return text
        for match in re_audio.finditer(text):
            filename = urllib.parse.quote(match.group(1).encode("utf-8"), safe="/:")
            text = text.replace(match.group(0), "")
            text += "<audio src=\"" + filename + "\" controls>"
        return text



#
# _sqlite3.py <Peter.Bienstman@UGent.be>
#

#
# Wrapper around sqlite3, to make it possible to easily swap an sqlite3
# backend with an APSW backend.
#

import os
import sys
import time
import sqlite3

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.libmnemosyne.utils import traceback_string, MnemosyneError


class _Sqlite3Cursor(object):

    def __init__(self, cursor):
        self.cursor = cursor

    def fetchone(self):
        return self.cursor.fetchone()

    def fetchall(self):
        return self.cursor.fetchall()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.cursor)


class _Sqlite3(Component):

    DEBUG = False

    def __init__(self, component_manager, path):
        Component.__init__(self, component_manager)
        self._cursor = None
        # Make sure we don't put a database on a network drive under Windows:
        # http://www.sqlite.org/lockingv3.html
        if sys.platform == "win32":  # pragma: no cover
            drive = os.path.splitdrive(path)[0]
            import ctypes
            if ctypes.windll.kernel32.GetDriveTypeW("%s\\" % drive) == 4:
                self.main_widget().show_error(_\
("Putting a database on a network drive is forbidden under Windows to avoid data corruption. Mnemosyne will now close."))
                sys.exit(-1)
        self.connection = sqlite3.connect(path)
        # http://www.mail-archive.com/sqlite-users@sqlite.org/msg34453.html
        self.connection.execute("pragma journal_mode = persist;")
        # Should only be used to speed up the test suite.
        if self.config()["asynchronous_database"] == True:
            self.connection.execute("pragma synchronous = off;")

    def executescript(self, script):
        if self.DEBUG:
            print(script)
            t = time.time()
        self.connection.executescript(script)
        if self.DEBUG:
            print(("took %.3f secs" % (time.time() - t)))

    def execute(self, sql, *args):
        if self.DEBUG:
            print((sql, args))
            t = time.time()
        try:
            self._cursor = self.connection.execute(sql, *args)
        except:
            raise MnemosyneError("SQL error: " + sql + " " + str(*args)
                + "\n" + traceback_string())
        if self.DEBUG:
            print(("took %.3f secs" % (time.time() - t)))
        return _Sqlite3Cursor(self._cursor)

    def executemany(self, sql, *args):
        if self.DEBUG:
            print((sql, args))
            t = time.time()
        self._cursor = self.connection.executemany(sql, *args)
        if self.DEBUG:
            print(("took %.3f secs" % (time.time() - t)))
        return _Sqlite3Cursor(self._cursor)

    def last_insert_rowid(self):
        return self._cursor.lastrowid

    def commit(self):
        return self.connection.commit()

    def close(self):
        del self._cursor
        return self.connection.close()



#
# no_gui_translator.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import GuiTranslator


class NoGuiTranslator(GuiTranslator):

    def set_translator(self, language):
        pass




#
# test_review_controller.py <Peter.Bienstman@UGent.be>
#

import os
import sys
import shutil

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne import Mnemosyne
from mnemosyne.libmnemosyne.criteria.default_criterion import DefaultCriterion
from mnemosyne.libmnemosyne.ui_components.review_widget import ReviewWidget

expected_scheduled_count = None

class MyReviewWidget(ReviewWidget):

    def update_status_bar_counters(self):
        scheduled_count, non_memorised_count, active_count = \
            self.review_controller().counters()
        if expected_scheduled_count is not None:
            assert scheduled_count == expected_scheduled_count

    def redraw_now(self):
        pass


class TestReviewController(MnemosyneTest):

    def setup(self):
        global expected_scheduled_count
        expected_scheduled_count = None
        self.initialise_data_dir()
        path = os.path.join(os.getcwd(), "..", "mnemosyne", "libmnemosyne",
                            "renderers")
        if path not in sys.path:
            sys.path.append(path)
        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.components.append(\
            ("mnemosyne.libmnemosyne.ui_components.main_widget", "MainWidget"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("test_review_controller", "MyReviewWidget")]
        self.mnemosyne.components.append(\
            ("mnemosyne.libmnemosyne.ui_components.dialogs", "EditCardDialog"))
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.review_controller().reset()

    def test_1(self):
        card_1 = None
        self.review_controller().reset()
        for i in range(10):
            fact_data = {"f": "question" + str(i),
                         "b": "answer" + str(i)}
            if i % 2:
                card_type = self.card_type_with_id("1")
            else:
                card_type = self.card_type_with_id("2")
            card = self.controller().create_new_cards(fact_data, card_type,
                    grade=4, tag_names=["default" + str(i)])[0]
            if i == 0:
                card_1 = card
                card.next_rep -= 1000 * 24 * 60 * 60
                self.database().update_card(card)
        self.review_controller().set_render_chain("default")
        self.review_controller().show_new_question()
        self.review_controller().reset_but_try_to_keep_current_card()
        self.review_controller().show_answer()
        assert self.review_controller().card == card_1
        assert self.review_controller().counters() == (1, 0, 15)
        self.review_controller().grade_answer(0)
        assert self.review_controller().counters() == (0, 1, 15)
        self.review_controller().grade_answer(2)
        assert self.review_controller().counters() == (0, 0, 15)
        self.review_controller().next_rep_string(0)
        self.review_controller().next_rep_string(1)
        self.review_controller().next_rep_string(2)

    def test_2(self):
        card_1 = None
        self.review_controller().reset()
        for i in range(10):
            fact_data = {"f": "question" + str(i),
                         "b": "answer" + str(i)}
            if i % 2:
                card_type = self.card_type_with_id("1")
            else:
                card_type = self.card_type_with_id("2")
            card = self.controller().create_new_cards(fact_data, card_type,
                    grade=4, tag_names=["default" + str(i)])[0]
            if i == 0:
                card_1 = card
                card.next_rep -= 1000 * 24 * 60 * 60
                self.database().update_card(card)
        self.review_controller().show_new_question()
        assert self.review_controller().card == card_1
        self.review_controller().reload_counters()
        assert self.review_controller().counters() == (1, 0, 15)
        self.review_controller().grade_answer(0)
        self.review_controller().reload_counters()
        assert self.review_controller().counters() == (0, 1, 15)
        self.review_controller().grade_answer(2)
        self.review_controller().reload_counters()
        assert self.review_controller().counters() == (0, 0, 15)

        self.mnemosyne.review_widget().set_grade_enabled(1, True)

    def test_reset_but_try_to_keep_current_card_turned_inactive(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "1", "b": "b"}
        card = self.controller().create_new_cards(fact_data, card_type,
                     grade=-1, tag_names=["forbidden"])[0]
        self.review_controller().show_new_question()
        assert self.review_controller().card == card

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("active")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        assert self.review_controller().card == card
        self.review_controller().reset_but_try_to_keep_current_card()
        assert self.review_controller().card is None

    def test_last_card(self):
        card_type = self.card_type_with_id("1")
        for data in ["1", "2", "3"]:
            fact_data = {"f": data, "b": data}
            self.controller().create_new_cards(fact_data, card_type, grade=-1, tag_names=[])
        self.review_controller().show_new_question()
        self.review_controller().show_answer()
        for i in range(5):
            self.review_controller().grade_answer(0)
            self.review_controller().show_answer()
        for i in range(2):
            self.review_controller().grade_answer(2)
            self.review_controller().show_answer()
        for i in range(6):
            self.review_controller().grade_answer(0)
            self.review_controller().show_answer()

    def test_counters(self):
        global expected_scheduled_count
        card_type = self.card_type_with_id("1")
        fact_data = {"f": '1', "b": '1'}
        card = self.controller().create_new_cards(fact_data, card_type, grade=5, tag_names=[])[0]
        card.next_rep = 0
        self.database().update_card(card)
        expected_scheduled_count = 1
        self.review_controller().show_new_question()
        assert self.review_controller().scheduled_count == 1
        assert self.review_controller().counters()[0] == 1
        self.review_controller().show_answer()
        expected_scheduled_count = 0
        self.review_controller().grade_answer(0)
        assert self.review_controller().scheduled_count == 0
        assert self.review_controller().counters()[0] == 0

    def test_counters_prefetch(self):
        global expected_scheduled_count
        card_type = self.card_type_with_id("1")
        for data in ['1', '2', '3', '4']:
            fact_data = {"f": data, "b": data}
            card = self.controller().create_new_cards(fact_data, card_type, grade=5, tag_names=[])[0]
            card.next_rep = 0
            self.database().update_card(card)
        expected_scheduled_count = 4
        self.review_controller().show_new_question()
        assert self.review_controller().scheduled_count == 4
        assert self.review_controller().counters()[0] == 4
        self.review_controller().show_answer()
        expected_scheduled_count = 3
        self.review_controller().grade_answer(3)
        assert self.review_controller().scheduled_count == 3
        assert self.review_controller().counters()[0] == 3



#
# convert_card_type_keys_dlg.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.pyqt_ui.ui_convert_card_type_keys_dlg import \
     Ui_ConvertCardTypeKeysDlg


class ConvertCardTypeKeysDlg(QtWidgets.QDialog, Ui_ConvertCardTypeKeysDlg):

    def __init__(self, old_card_type, new_card_type, correspondence,
                 check_required_fact_keys=True, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        self.old_card_type = old_card_type
        self.new_card_type = new_card_type
        self.correspondence = correspondence
        self.check_required_fact_keys = check_required_fact_keys
        self.comboboxes = {}
        index = 1
        for old_fact_key, old_fact_key_name in \
            old_card_type.fact_keys_and_names:
            label = QtWidgets.QLabel(self)
            label.setText(_(old_fact_key_name) + ":")
            font = QtGui.QFont()
            font.setWeight(50)
            font.setBold(False)
            label.setFont(font)
            self.gridLayout.addWidget(label, index, 0, 1, 1)
            combobox = QtWidgets.QComboBox(self)
            for new_fact_key, new_key_name in \
                new_card_type.fact_keys_and_names:
                combobox.addItem(_(new_key_name))
            combobox.addItem(_("<none>"))
            combobox.setCurrentIndex(combobox.count()-1)
            self.gridLayout.addWidget(combobox, index, 1, 1, 1)
            self.comboboxes[old_fact_key] = combobox
            index += 1
            combobox.currentIndexChanged.connect(self.combobox_updated)

    def combobox_updated(self):
        self.ok_button.setEnabled(False)
        self.correspondence.clear()
        for old_fact_key, old_fact_key_name in \
            self.old_card_type.fact_keys_and_names:
            new_fact_key_name = \
                self.comboboxes[old_fact_key].currentText()
            if new_fact_key_name != _("<none>"):
                self.ok_button.setEnabled(True)
                new_fact_key = \
                     self.new_card_type.fact_key_with_name(new_fact_key_name)
                if new_fact_key in self.correspondence.values():
                    QtWidgets.QMessageBox.critical(self, _("Mnemosyne"),
                        _("No duplicate in new fact keys allowed."))
                    self.ok_button.setEnabled(False)
                    return
                self.correspondence[old_fact_key] = new_fact_key
        if self.check_required_fact_keys:
            for fact_key in self.new_card_type.required_fact_keys:
                if fact_key not in self.correspondence.values():
                    self.ok_button.setEnabled(False)
                    if len(self.correspondence) == \
                       len(self.old_card_type.fact_keys()):
                        QtWidgets.QMessageBox.critical(self, _("Mnemosyne"),
                            _("A required field is missing."))
                    return

    def accept(self):
        self.correspondence.clear()
        for old_fact_key, old_fact_key_name in \
            self.old_card_type.fact_keys_and_names:
            new_fact_key_name = \
                self.comboboxes[old_fact_key].currentText()
            if new_fact_key_name != _("<none>"):
                new_fact_key = \
                     self.new_card_type.fact_key_with_name(new_fact_key_name)
                self.correspondence[old_fact_key] = new_fact_key
        QtWidgets.QDialog.accept(self)
